#!/usr/bin/env python3

"""
Inspect checkpoint
"""

from __future__ import annotations

import os
import argparse
import numpy

import _setup_returnn_env  # noqa
from torch_inspect_checkpoint import parse_numpy_printoption, print_object, _print_key_value  # noqa
from returnn.config import Config, set_global_config
from returnn.log import log
from returnn.torch.engine import Engine


def main():
    """main"""
    print(f"{os.path.basename(__file__)}: {__doc__.strip()}")

    # First set PyTorch default print options. This might get overwritten by the args below.
    numpy.set_printoptions(precision=4, linewidth=80)

    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument("returnn_config")
    arg_parser.add_argument("--cwd")
    arg_parser.add_argument("--key", type=str, default="", help="Name of the tensor or object to inspect")
    arg_parser.add_argument("--all_tensors", action="store_true", help="If True, print the values of all the tensors.")
    arg_parser.add_argument("--stats_only", action="store_true")
    arg_parser.add_argument(
        "--printoptions",
        nargs="*",
        type=parse_numpy_printoption,
        help="Argument for numpy.set_printoptions(), in the form 'k=v'.",
    )
    arg_parser.add_argument("--device", default="cpu")
    args = arg_parser.parse_args()

    if args.cwd:
        print("* Change working dir:", args.cwd)
        os.chdir(args.cwd)

    log.initialize(verbosity=[5])
    config = Config()
    print("* Load config:", args.returnn_config)
    config.load_file(args.returnn_config)
    set_global_config(config)
    for k in [
        "train",
        "dev",
        "eval",
        "eval_datasets",
        "torch_amp",
        "grad_scaler",
        "torch_distributed",
        "learning_rate_control",
        "learning_rate_file",
    ]:
        config.typed_dict.pop(k, None)
    config.set("device", args.device)

    print("* Setup RETURNN engine")
    engine = Engine(config=config)
    print("* Load model and optimizer state")
    engine.init_train_from_config()
    model = engine.get_pt_model()
    assert model is not None, "No model loaded?"
    opt = engine.get_pt_optimizer()
    assert opt is not None, "No optimizer loaded?"
    print("* Loaded.")

    if args.key:
        obj = model.get_parameter(args.key)
        print(f"{args.key}:")
        print_object(obj, stats_only=args.stats_only)
        print("Optimizer state:")
        if obj in opt.state:
            print_object(opt.state[obj], stats_only=args.stats_only)
        else:
            print("(None)")
    else:
        for name, param in model.named_parameters():
            _print_key_value(name, param, print_all_tensors=args.all_tensors, stats_only=args.stats_only)
            if param in opt.state:
                print("  Optimizer state:")
                print_object(
                    opt.state[param], prefix="  ", print_all_tensors=args.all_tensors, stats_only=args.stats_only
                )
            else:
                print("  Optimizer state: (None)")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

"""
File cache utility.
"""

from __future__ import annotations

import _setup_returnn_env  # noqa
import sys
import os
import shutil
import argparse
from returnn.util import file_cache
from returnn.util.basic import get_temp_dir, human_bytes_size


def main():
    """
    Main entry.
    """
    argparser = argparse.ArgumentParser(description="Dump something from dataset.")
    argparser.add_argument("--temp-dir", help="Temp directory to use.", default=None)
    argparser.add_argument("--cache-dir", help="Cache directory to use.", default=None)
    argparser.add_argument("--cleanup-require-free-space", type=_parse_size, default=0)
    argparser.add_argument(
        "--cleanup-files-always-older-than-days", type=float, help="Cleanup files always older than this many days."
    )
    argparser.add_argument(
        "--cleanup-files-wanted-older-than-days", type=float, help="Cleanup files wanted older than this many days."
    )
    argparser.add_argument(
        "--cleanup-disk-usage-wanted-free-ratio", type=float, help="Cleanup to reach this free disk space ratio."
    )
    argparser.add_argument(
        "--cleanup-disk-usage-wanted-multiplier", type=float, help="Cleanup to free this multiplier of required space."
    )

    argparser.add_argument("action", help="Action to perform.", choices=["info", "cleanup"])
    args = argparser.parse_args()

    if args.temp_dir:
        os.environ["TMPDIR"] = args.temp_dir
    # logic from get_temp_dir:
    for envname in ["TMPDIR", "TEMP", "TMP"]:
        dirname = os.getenv(envname)
        print(f"env {envname}: {dirname}")
        if dirname:
            break
    print("temp dir:", get_temp_dir(with_username=False))

    cache = file_cache.FileCache(
        **({"cache_directory": args.cache_dir} if args.cache_dir else {}),
        **(
            {"cleanup_files_always_older_than_days": args.cleanup_files_always_older_than_days}
            if args.cleanup_files_always_older_than_days is not None
            else {}
        ),
        **(
            {"cleanup_files_wanted_older_than_days": args.cleanup_files_wanted_older_than_days}
            if args.cleanup_files_wanted_older_than_days is not None
            else {}
        ),
        **(
            {"cleanup_disk_usage_wanted_free_ratio": args.cleanup_disk_usage_wanted_free_ratio}
            if args.cleanup_disk_usage_wanted_free_ratio is not None
            else {}
        ),
        **(
            {"cleanup_disk_usage_wanted_multiplier": args.cleanup_disk_usage_wanted_multiplier}
            if args.cleanup_disk_usage_wanted_multiplier is not None
            else {}
        ),
    )
    print("cache dir:", cache.cache_directory)
    if not os.path.exists(cache.cache_directory):
        print("Cache dir does not exist.")
        sys.exit(1)

    disk_usage = shutil.disk_usage(cache.cache_directory)
    print(
        f"Cache dir, disk usage: total {human_bytes_size(disk_usage.total)},"
        f" used {human_bytes_size(disk_usage.used)},"
        f" free {human_bytes_size(disk_usage.free)}"
    )

    if args.action == "cleanup":
        print("Cleanup.")
        print(f"Require free space: {human_bytes_size(args.cleanup_require_free_space)}")
        res = cache.cleanup(need_at_least_free_space_size=args.cleanup_require_free_space)
        print(res)


def _parse_size(size_str: str) -> int:
    """
    Parse size string like "10G" or "500M" to integer bytes.
    """
    size_str = size_str.strip().upper()
    if size_str.endswith("G"):
        return int(float(size_str[:-1]) * (1024**3))
    if size_str.endswith("M"):
        return int(float(size_str[:-1]) * (1024**2))
    if size_str.endswith("K"):
        return int(float(size_str[:-1]) * 1024)
    return int(size_str)


if __name__ == "__main__":
    main()



"""
This module provides some helpers to support earlier versions of TF 1 and also TF 2.
In all cases, we use graph-mode execution, i.e. we will disable eager-mode.

For an initial draft about TF 2 support, see here:
https://github.com/rwth-i6/returnn/issues/283

Also see this for a similar (but different) approach:
https://github.com/tensorflow/lingvo/blob/master/lingvo/compat.py
"""

import tensorflow as tf
import os as _os

if not getattr(tf, "compat", None) or not getattr(tf.compat, "v2", None):
    v1 = tf
    v2 = None
else:
    # PyCharm type-inference will take the latest reference,
    # so this `else` branch should lead us to a valid reference for "modern" TF versions (TF >=1.14, or TF 2).
    v1 = tf.compat.v1
    v2 = tf.compat.v2

if v2 and tf.__version__.startswith("2."):
    tf.compat.v1.disable_eager_execution()
    tf.compat.v1.disable_v2_tensorshape()
    # There are still issues with control flow v2,
    # so we do not enable it yet by default.
    if _os.environ.get("TF_ENABLE_COND_V2", "") == "1":
        # https://github.com/tensorflow/tensorflow/issues/54458
        tf.compat.v1.experimental.output_all_intermediates(True)
    else:
        tf.compat.v1.disable_control_flow_v2()
    # There is also disable_v2_behavior() which disables all the TF2 behavior,
    # including control flow, eager execution, etc.,
    # but we want to make use of TF2 features as much as possible.


def executing_eagerly():
    """
    :return: True if we are currently executing eagerly.
    :rtype: bool
    """
    if hasattr(tf, "executing_eagerly"):
        return tf.executing_eagerly()
    return False



"""
Some mathematical functions, in pure NumPy.
"""

from __future__ import annotations
from typing import Union, Optional, Any, Sequence, Dict
import numpy
import hashlib
import bisect


def ceil_div(a: int, b: int) -> int:
    """ceil(a / b)"""
    return -(-a // b)


def next_power_of_two(n: int) -> int:
    """next power of two, >= n"""
    return 2 ** (int(n - 1).bit_length())


class PiecewiseLinear:
    """
    Piecewise linear function.
    (Basically wraps ``numpy.interp``.)
    """

    def __init__(
        self,
        values: Dict[Union[int, float], Union[int, float]],
        *,
        kw_name: Optional[str] = None,
        ignore_other_kwargs: bool = False,
    ):
        """
        :param values: dict x -> y. Everything between the x values is linearly interpolated.
            Everything outside is repeated from the nearest x value.
        :param kw_name: keyword argument name to use in the __call__. Other keyword arguments are ignored.
        :param ignore_other_kwargs: if True, ignore other keyword arguments in the __call__.
        """
        self._sorted_items = sorted(values.items())
        self._sorted_keys = numpy.array([x for x, _ in self._sorted_items])
        self._sorted_values = numpy.array([y for _, y in self._sorted_items])
        self._kw_name = kw_name
        self._ignore_other_kwargs = ignore_other_kwargs

    def __getstate__(self):
        # Note: I was implementing __getnewargs_ex__, but we cannot use this because of this Sisyphus bug:
        # https://github.com/rwth-i6/sisyphus/issues/215
        kwargs = {"values": dict(self._sorted_items)}
        if self._kw_name is not None:
            kwargs["kw_name"] = self._kw_name
        if self._ignore_other_kwargs:
            kwargs["ignore_other_kwargs"] = True
        return kwargs

    def __setstate__(self, state):
        self.__init__(**state)

    def __repr__(self) -> str:
        kwargs = self.__getstate__()
        values = kwargs.pop("values")
        all_args_str = ", ".join([repr(values)] + [f"{k}={v!r}" for k, v in kwargs.items()])
        return f"{self.__class__.__name__}({all_args_str})"

    def __call__(self, *args: Union[int, float], **kwargs) -> Union[int, float]:
        if self._kw_name:
            if args:
                raise TypeError(f"{self}: Expected zero positional arguments, got {args!r}")
            x = kwargs.pop(self._kw_name, None)
        else:
            if len(args) != 1:
                raise TypeError(f"{self}: Expected one positional argument, got {args!r}")
            x = args[0]
        if not self._ignore_other_kwargs:
            if kwargs:
                raise TypeError(f"{self}: Unexpected keyword arguments: {kwargs!r}")
        assert x is not None
        steps = self._sorted_keys
        values = self._sorted_values
        return numpy.interp(x, steps, values)


class StepFunction:
    """
    Step function. (Piecewise constant function.)

    It will find the nearest x value and return the corresponding y value.
    The corresponding y value can be anything here, not just numbers.
    """

    def __init__(
        self,
        values: Dict[Union[int, float], Any],
        *,
        kw_name: Optional[str] = None,
        ignore_other_kwargs: bool = False,
    ):
        """
        :param values: dict x -> y. Everything between the x values is the value of the nearest x.
        :param kw_name: keyword argument name to use in the __call__. Other keyword arguments are ignored.
        :param ignore_other_kwargs: if True, ignore other keyword arguments in the __call__.
        """
        assert values, f"{self}: No values provided"
        self._values = values
        self._sorted_keys = sorted(values.keys())
        self._kw_name = kw_name
        self._ignore_other_kwargs = ignore_other_kwargs

    def __getstate__(self):
        # Note: I was implementing __getnewargs_ex__, but we cannot use this because of this Sisyphus bug:
        # https://github.com/rwth-i6/sisyphus/issues/215
        kwargs = {"values": self._values}
        if self._kw_name is not None:
            kwargs["kw_name"] = self._kw_name
        if self._ignore_other_kwargs:
            kwargs["ignore_other_kwargs"] = True
        return kwargs

    def __setstate__(self, state):
        self.__init__(**state)

    def __repr__(self) -> str:
        kwargs = self.__getstate__()
        values = kwargs.pop("values")
        all_args_str = ", ".join([repr(values)] + [f"{k}={v!r}" for k, v in kwargs.items()])
        return f"{self.__class__.__name__}({all_args_str})"

    def __call__(self, *args: Union[int, float], **kwargs) -> Union[int, float]:
        if self._kw_name:
            if args:
                raise TypeError(f"{self}: Expected zero positional arguments, got {args!r}")
            x = kwargs.pop(self._kw_name, None)
        else:
            if len(args) != 1:
                raise TypeError(f"{self}: Expected one positional argument, got {args!r}")
            x = args[0]
        if not self._ignore_other_kwargs:
            if kwargs:
                raise TypeError(f"{self}: Unexpected keyword arguments: {kwargs!r}")
        assert x is not None
        keys = self._sorted_keys
        assert keys, f"{self}: No keys provided"
        i = bisect.bisect_left(keys, x)
        # keys[:i] have e < x, and all e in keys[i:] have e >= x
        # keys[i] is the smallest e >= x
        if i >= len(keys) or (i > 0 and (x - keys[i - 1]) <= (keys[i] - x)):
            i -= 1
        return self._values[keys[i]]


def simplify_and_format_number(n: Union[int, float]) -> str:
    """Format a number, removing trailing zeros and the decimal point if it is an integer"""
    if isinstance(n, (int, float)):
        return str(n).rstrip("0").rstrip(".")
    else:
        raise TypeError(f"Expected int or float, got {n!r} type {type(n)}")


def merge_random_seeds(data_sources: Sequence[int], *, num_bytes: int = 4, signed: bool = False) -> int:
    """
    :param data_sources: A list of integers. We expect that they are all representable as 64-bit signed integers.
    :param num_bytes: for the output seed.
    :param signed: whether the output seed should be signed.
    :return: A num_bytes*8-bit integer seed, deterministically derived from the input data.
    """
    # Convert each integer to bytes and concatenate them
    combined = b"".join(int(source).to_bytes(8, "big", signed=True) for source in data_sources)
    # Use SHA-256 to hash the combined bytes
    hash_digest = hashlib.sha256(combined).digest()
    # Convert the hash digest to an integer seed
    seed = int.from_bytes(hash_digest[:num_bytes], "big", signed=signed)
    return seed



#!/usr/bin/env python3

"""
This script imports a t2t model into Returnn and compares the activations.
It currently assumes a specific Returnn network topology with specific layer names.
"""

from __future__ import annotations

import os
import sys
from pprint import pprint
import tensorflow as tf

import numpy

# from numpy.testing import assert_almost_equal

my_dir = os.path.dirname(os.path.abspath(__file__))
returnn_dir = os.path.dirname(my_dir)
sys.path.insert(0, returnn_dir)

import returnn.util.basic as util
from returnn.util import better_exchook
import returnn.__main__ as rnn
import returnn.tf.compat as tf_compat
from returnn.tf.network import TFNetwork
from returnn.tf.layers.basic import SourceLayer, LayerBase, LinearLayer


import tensor2tensor
import tensor2tensor.models
from tensor2tensor.bin import t2t_trainer

# from tensor2tensor.data_generators import problem  # pylint: disable=unused-import
from tensor2tensor.data_generators import text_encoder

# from tensor2tensor.utils import decoding
from tensor2tensor.utils import registry
from tensor2tensor.utils import trainer_lib
from tensor2tensor.utils import usr_dir


import ipdb


# T2T_MODEL_DIR = "/work/smt3/schamper/sandbox/t2t-multi30k_2layers"
T2T_MODEL_DIR = "/work/smt3/schamper/sandbox/out_model_2layer_512"
FLAGS_data_dir = T2T_MODEL_DIR + "/datadir"  # the data created by the self defined problem
FLAGS_output_dir = T2T_MODEL_DIR + "/out_model"  # the trained t2tmodel
FLAGS_score_file = T2T_MODEL_DIR + "/to_score"  # The data we want to test on

# model was trained with self defined problem 'translate_tmp' to handle own data paths
# but importing works also with compatible standard problem
FLAGS_problem = "translate_ende_wmt8k"
FLAGS_model = "transformer"
FLAGS_hparams_set = "transformer_base_single_gpu"
# FLAGS_hparams = "num_hidden_layers=2" # default is empty
FLAGS_hparams = "num_hidden_layers=2,hidden_size=256,shared_embedding_and_softmax_weights=False,symbol_modality_num_shards=1,conv_first_kernel=0,use_target_space_embedding=False,filter_size=512"


def create_t2t_hparams():
    return trainer_lib.create_hparams(
        FLAGS_hparams_set, FLAGS_hparams, data_dir=os.path.expanduser(FLAGS_data_dir), problem_name=FLAGS_problem
    )


def t2t_score_file(filename):
    """
    Score each line in a file and return the scores.

    :param str filename: T2T checkpoint
    """
    # Prepare model.
    hparams = create_t2t_hparams()
    encoders = registry.problem(FLAGS_problem).feature_encoders(FLAGS_data_dir)

    # Prepare features for feeding into the model.
    inputs_ph = tf_compat.v1.placeholder(dtype=tf.int32, shape=(None, None))  # Just length dimension.
    targets_ph = tf_compat.v1.placeholder(dtype=tf.int32, shape=(None, None))  # Just length dimension.

    features = {
        "inputs": inputs_ph,
        "targets": targets_ph,
    }

    # Prepare the model and the graph when model runs on features.
    model = registry.model(FLAGS_model)(hparams, tf.estimator.ModeKeys.EVAL)
    assert isinstance(model, tensor2tensor.models.transformer.Transformer)
    #       final_output: tensor of logits with shape [batch_size, O, P, body_output_size.
    #       losses: either single loss as a scalar, a list, a tensor (to be averaged)
    #               or a dictionary of losses.
    final_output, losses = model(features)
    assert isinstance(losses, dict)
    saver = tf_compat.v1.train.Saver()

    sess = tf_compat.v1.Session()
    # Load weights from checkpoint.
    ckpts = tf.train.get_checkpoint_state(FLAGS_output_dir)
    ckpt = ckpts.model_checkpoint_path
    saver.restore(sess, ckpt)

    # writer = tf.compat.v1.summary.FileWriter('logs', sess.graph)

    # writer.close()

    # Run on each line.
    results = []
    for line in open(filename):
        tab_split = line.split("\t")
        if len(tab_split) > 2:
            raise ValueError("Each line must have at most one tab separator.")
        assert len(tab_split) == 2
        targets = tab_split[1].strip()
        inputs = tab_split[0].strip()
        # Run encoders and append EOS symbol.
        targets_numpy = encoders["targets"].encode(targets) + [text_encoder.EOS_ID]
        inputs_numpy = encoders["inputs"].encode(inputs) + [text_encoder.EOS_ID]
        # Prepare the feed.
        feed = {inputs_ph: [inputs_numpy], targets_ph: [targets_numpy]}

        np_res = sess.run({"losses": losses, "final_output": final_output}, feed_dict=feed)
        pprint(np_res)

        tvars = tf_compat.v1.trainable_variables()

        print("t2t inputs_ph:", inputs_ph, inputs_numpy)
        print("t2t targets_ph:", targets_ph, targets_numpy)

        return sess, tvars, inputs_ph, targets_ph, losses


FFDim = 512
EncKeyTotalDim = 256
AttNumHeads = 8
EncKeyPerHeadDim = EncKeyTotalDim // AttNumHeads
EncValueTotalDim = 256
EncValuePerHeadDim = EncValueTotalDim // AttNumHeads


# consider: /work/smt3/bahar/debug/returnn/t2t-test/02.07-2018-test/returnn/config-new.py


def add_trafo_enc_layer(d, inp, output):
    d[output + "_self_att_laynorm"] = {"class": "layer_norm", "from": [inp]}
    d[output + "_self_att_"] = {
        "class": "self_attention",
        "num_heads": AttNumHeads,
        "total_key_dim": EncKeyTotalDim,
        "n_out": EncValueTotalDim,
        "from": [output + "_self_att_laynorm"],
        "attention_left_only": False,
    }
    d[output + "_self_att_lin"] = {
        "class": "linear",
        "activation": None,
        "with_bias": False,
        "from": [output + "_self_att_"],
        "n_out": EncValueTotalDim,
    }
    d[output + "_self_att_drop"] = {"class": "dropout", "from": [output + "_self_att_lin"], "dropout": 0.1}
    d[output + "_self_att_out"] = {
        "class": "combine",
        "kind": "add",
        "from": [inp, output + "_self_att_drop"],
        "n_out": EncValueTotalDim,
    }
    #####
    d[output + "_ff_laynorm"] = {"class": "layer_norm", "from": [output + "_self_att_out"]}
    d[output + "_ff_conv1"] = {
        "class": "linear",
        "activation": "relu",
        "with_bias": True,
        "from": [output + "_ff_laynorm"],
        "n_out": FFDim,
    }
    d[output + "_ff_conv2"] = {
        "class": "linear",
        "activation": None,
        "with_bias": True,
        "from": [output + "_ff_conv1"],
        "n_out": EncValueTotalDim,
    }
    d[output + "_ff_drop"] = {"class": "dropout", "from": [output + "_ff_conv2"], "dropout": 0.1}
    d[output + "_ff_out"] = {
        "class": "combine",
        "kind": "add",
        "from": [output + "_self_att_out", output + "_ff_drop"],
        "n_out": EncValueTotalDim,
    }
    d[output] = {"class": "copy", "from": [output + "_ff_out"]}


def add_trafo_dec_layer(db, d, inp, output):
    d[output + "_self_att_laynorm"] = {"class": "layer_norm", "from": [inp]}
    d[output + "_self_att_"] = {
        "class": "self_attention",
        "num_heads": AttNumHeads,
        "total_key_dim": EncKeyTotalDim,
        "n_out": EncValueTotalDim,
        "from": [output + "_self_att_laynorm"],
        "attention_left_only": True,
    }
    d[output + "_self_att_lin"] = {
        "class": "linear",
        "activation": None,
        "with_bias": False,
        "from": [output + "_self_att_"],
        "n_out": EncValueTotalDim,
    }
    d[output + "_self_att_drop"] = {"class": "dropout", "from": [output + "_self_att_lin"], "dropout": 0.1}
    d[output + "_self_att_out"] = {
        "class": "combine",
        "kind": "add",
        "from": [inp, output + "_self_att_drop"],
        "n_out": EncValueTotalDim,
    }
    #####
    d[output + "_att_laynorm"] = {"class": "layer_norm", "from": [output + "_self_att_out"]}
    d[output + "_att_query0"] = {
        "class": "linear",
        "activation": None,
        "with_bias": False,
        "from": [output + "_att_laynorm"],
        "n_out": EncValueTotalDim,
    }
    d[output + "_att_query"] = {
        "class": "split_dims",
        "axis": "F",
        "dims": (AttNumHeads, EncKeyPerHeadDim),
        "from": [output + "_att_query0"],
    }  # (B, H, D/H)
    db[output + "_att_key0"] = {
        "class": "linear",
        "activation": None,
        "with_bias": False,
        "from": ["encoder"],
        "n_out": EncKeyTotalDim,
    }  # (B, enc-T, D)
    db[output + "_att_value0"] = {
        "class": "linear",
        "activation": None,
        "with_bias": False,
        "from": ["encoder"],
        "n_out": EncValueTotalDim,
    }
    db[output + "_att_key"] = {
        "class": "split_dims",
        "axis": "F",
        "dims": (AttNumHeads, EncKeyPerHeadDim),
        "from": [output + "_att_key0"],
    }  # (B, enc-T, H, D/H)
    db[output + "_att_value"] = {
        "class": "split_dims",
        "axis": "F",
        "dims": (AttNumHeads, EncValuePerHeadDim),
        "from": [output + "_att_value0"],
    }  # (B, enc-T, H, D'/H)
    d[output + "_att_energy"] = {
        "class": "dot",
        "red1": "static:-1",
        "red2": "static:-1",
        "var1": "T",
        "var2": "T?",
        "from": ["base:" + output + "_att_key", output + "_att_query"],
    }  # (B, H, enc-T, 1)
    d[output + "_att_weights"] = {
        "class": "softmax_over_spatial",
        "from": [output + "_att_energy"],
        "energy_factor": EncKeyPerHeadDim**-0.5,
    }  # (B, enc-T, H, 1)
    d[output + "_att0"] = {
        "class": "generic_attention",
        "weights": output + "_att_weights",
        "base": "base:" + output + "_att_value",
    }  # (B, H, V)
    d[output + "_att_"] = {"class": "merge_dims", "axes": "static", "from": [output + "_att0"]}  # (B, H*V) except_batch
    d[output + "_att_lin"] = {
        "class": "linear",
        "activation": None,
        "with_bias": False,
        "from": [output + "_att_"],
        "n_out": EncValueTotalDim,
    }
    d[output + "_att_drop"] = {"class": "dropout", "from": [output + "_att_lin"], "dropout": 0.1}
    d[output + "_att_out"] = {
        "class": "combine",
        "kind": "add",
        "from": [output + "_self_att_out", output + "_att_drop"],
        "n_out": EncValueTotalDim,
    }
    #####
    d[output + "_ff_laynorm"] = {"class": "layer_norm", "from": [output + "_att_out"]}
    d[output + "_ff_conv1"] = {
        "class": "linear",
        "activation": "relu",
        "with_bias": True,
        "from": [output + "_ff_laynorm"],
        "n_out": FFDim,
    }
    d[output + "_ff_conv2"] = {
        "class": "linear",
        "activation": None,
        "with_bias": True,
        "from": [output + "_ff_conv1"],
        "n_out": EncValueTotalDim,
    }
    d[output + "_ff_drop"] = {"class": "dropout", "from": [output + "_ff_conv2"], "dropout": 0.1}
    d[output + "_ff_out"] = {
        "class": "combine",
        "kind": "add",
        "from": [output + "_att_out", output + "_ff_drop"],
        "n_out": EncValueTotalDim,
    }
    d[output] = {"class": "copy", "from": [output + "_ff_out"]}


# network
# (also defined by num_inputs & num_outputs)
returnn_network = {
    "source_embed_raw": {"class": "linear", "activation": None, "with_bias": False, "n_out": EncValueTotalDim},
    "source_embed_weighted": {
        "class": "eval",
        "from": ["source_embed_raw"],
        "eval": "source(0) * (%i**0.5)" % EncValueTotalDim,
    },
    "source_embed_with_pos": {
        "class": "positional_encoding",
        "add_to_input": True,
        "from": ["source_embed_weighted"],
        "dropout": 0.1,
    },
    "source_embed": {"class": "copy", "from": ["source_embed_with_pos"]},
    ## trafo layer added later
    "encoder": {"class": "layer_norm", "from": ["enc_N"]},
    "output": {
        "class": "rec",
        "from": [],
        "unit": {
            "output": {"class": "choice", "target": "classes", "beam_size": 12, "from": ["output_prob"]},
            "end": {"class": "compare", "from": ["output"], "value": 0},
            "target_embed_raw": {
                "class": "linear",
                "activation": None,
                "with_bias": False,
                "from": ["output"],
                "n_out": EncValueTotalDim,
                "initial_output": 0,
            },  # there seems to be no <s> in t2t, they seem to use just the zero vector
            "target_embed_weighted": {
                "class": "eval",
                "from": ["prev:target_embed_raw"],
                "eval": "source(0) * (%i**0.5)" % EncValueTotalDim,
            },
            "target_embed_with_pos": {
                "class": "positional_encoding",
                "add_to_input": True,
                "from": ["target_embed_weighted"],
            },
            "target_embed": {"class": "dropout", "from": ["target_embed_with_pos"], "dropout": 0.1},
            ## trafo layer added later
            # ToDo: Add last linear layer???
            "decoder": {"class": "layer_norm", "from": ["dec_N"]},
            "output_prob": {
                "class": "softmax",
                "from": ["decoder"],
                "dropout": 0.0,
                "target": "classes",
                "loss": "ce",
                "loss_opts": {"label_smoothing": 0.1},
                "with_bias": False,
            },
        },
        "target": "classes",
        "max_seq_len": "max_len_from('base:encoder') * 3",
    },
}

add_trafo_enc_layer(returnn_network, "source_embed", "enc_1")
add_trafo_enc_layer(returnn_network, "enc_1", "enc_N")
add_trafo_dec_layer(returnn_network, returnn_network["output"]["unit"], "target_embed", "dec_1")
add_trafo_dec_layer(returnn_network, returnn_network["output"]["unit"], "dec_1", "dec_N")


num_outputs = {"classes": [6115, 1], "data": [6115, 1]}
num_inputs = num_outputs["data"][0]


def main():
    print("#####################################################")
    print("Loading t2t model + scoring")
    t2t_sess, t2t_tvars, t2t_inputs_ph, t2t_targets_ph, t2t_losses = t2t_score_file(FLAGS_score_file)

    print("#####################################################")
    print("Loading returnn config")

    rnn.init(
        config_updates={
            "optimize_move_layers_out": True,
            "use_tensorflow": True,
            "num_outputs": num_outputs,
            "num_inputs": num_inputs,
            "task": "nop",
            "log": None,
            "device": "cpu",
            "network": returnn_network,
            "debug_print_layer_output_template": True,
            "debug_add_check_numerics_on_output": False,
        },
        extra_greeting="Import t2t model.",
    )
    assert util.BackendEngine.is_tensorflow_selected()
    config = rnn.config

    rnn.engine.init_train_from_config(config=config)
    network = rnn.engine.network
    assert isinstance(network, TFNetwork)

    print("t2t network model params:")
    t2t_params = {}  # type: dict[str,tf.Variable]
    t2t_total_num_params = 0
    for v in t2t_tvars:
        key = v.name[:-2]
        t2t_params[key] = v
        print("  %s: %s, %s" % (key, v.shape, v.dtype.base_dtype.name))
        t2t_total_num_params += numpy.prod(v.shape.as_list())
    print("t2t total num params: %i" % t2t_total_num_params)

    print("Our network model params:")
    our_params = {}  # type: dict[str,tf.Variable]
    our_total_num_params = 0
    for v in network.get_params_list():
        key = v.name[:-2]
        our_params[key] = v
        print("  %s: %s, %s" % (key, v.shape, v.dtype.base_dtype.name))
        our_total_num_params += numpy.prod(v.shape.as_list())
    print("Our total num params: %i" % our_total_num_params)

    print("Loading t2t params into our network:")
    for ret_var_name, ret_var in our_params.items():
        if ret_var_name in ret_to_t2t:
            t2t_var_names = ret_to_t2t[ret_var_name]
            # in return QKV params are concatenated into one tensor
            if isinstance(t2t_var_names, tuple):
                # params_np = numpy.concatenate([t2t_params[var_name].eval(t2t_sess) for var_name in t2t_var_names], axis=1) # Not enough...
                # More complex stacking necessary: head1 Q, head1 K, head1 V,   head2 Q, head2 K, head2 V, ...
                q_np = t2t_params[t2t_var_names[0]].eval(t2t_sess)
                k_np = t2t_params[t2t_var_names[1]].eval(t2t_sess)
                v_np = t2t_params[t2t_var_names[2]].eval(t2t_sess)
                qkv_dim_total = 2 * EncKeyTotalDim + EncValueTotalDim
                params_np = numpy.empty((EncValueTotalDim, qkv_dim_total), dtype=q_np.dtype)
                qkv_dim_per_head = qkv_dim_total // AttNumHeads
                for i in range(EncKeyPerHeadDim):
                    params_np[:, i::qkv_dim_per_head] = q_np[:, i::EncKeyPerHeadDim]
                for i in range(EncKeyPerHeadDim):
                    params_np[:, i + EncKeyPerHeadDim :: qkv_dim_per_head] = k_np[:, i::EncKeyPerHeadDim]
                for i in range(EncValuePerHeadDim):
                    params_np[:, i + 2 * EncKeyPerHeadDim :: qkv_dim_per_head] = v_np[:, i::EncValuePerHeadDim]
            else:
                t2t_var = t2t_params[t2t_var_names]
                params_np = t2t_var.eval(t2t_sess)
            if ret_var_name in ["output/rec/output_prob/W"]:  # ToDo: Something else to transpose?
                params_np = params_np.transpose()
            if ret_var_name in ["source_embed_raw/W", "output/rec/target_embed_raw/W"]:
                # params_np = params_np * (EncValueTotalDim**0.5) # ToDo: Only because of weight-tying?
                print("loading %s with * (EncValueTotalDim**0.5) or doing it in config" % ret_var.name)
            ret_var.load(params_np, rnn.engine.tf_session)
            print("loaded %s" % ret_var.name)
        else:
            print("skpipped over %s" % ret_var.name)

    ret_ph_train = rnn.engine.tf_session.graph.get_tensor_by_name("global_tensor_train_flag/train_flag:0")
    ret_ph_data = rnn.engine.tf_session.graph.get_tensor_by_name("extern_data/placeholders/data/data:0")
    ret_ph_data_dim = rnn.engine.tf_session.graph.get_tensor_by_name("extern_data/placeholders/data/data_dim0_size:0")
    ret_ph_classes = rnn.engine.tf_session.graph.get_tensor_by_name("extern_data/placeholders/classes/classes:0")
    ret_ph_classes_dim = rnn.engine.tf_session.graph.get_tensor_by_name(
        "extern_data/placeholders/classes/classes_dim0_size:0"
    )

    # ret_feed = {ret_ph_train: True, ret_ph_data: [[11, 78, 42, 670, 2415, 2, 134, 2, 61, 522, 2, 847, 2, 3353, 15, 33, 2534, 1], [3,6]], ret_ph_data_dim: [18, 2],
    #            ret_ph_classes: [[4, 60, 18, 46, 26, 2937, 520, 2, 1317, 2, 10, 642, 4, 639, 1], [2,5]], ret_ph_classes_dim:[14, 2]}

    # src = [[78, 1,0], [2, 134, 1]]; src_lens = [2,3]; trg = [[4, 60, 1], [639, 1, 0]]; trg_lens = [3,2]; ret_feed = {ret_ph_train: False, ret_ph_data: src, ret_ph_data_dim: src_lens, ret_ph_classes: trg, ret_ph_classes_dim: trg_lens}; t2t_feed = {t2t_inputs_ph: src, t2t_targets_ph: trg}

    src = [[2, 134, 1]]
    src_lens = [3]
    trg = [[4, 60, 1]]
    trg_lens = [3]
    ret_feed = {
        ret_ph_train: False,
        ret_ph_data: src,
        ret_ph_data_dim: src_lens,
        ret_ph_classes: trg,
        ret_ph_classes_dim: trg_lens,
    }
    t2t_feed = {t2t_inputs_ph: src, t2t_targets_ph: trg}

    compare_acts(network, t2t_sess, ret_feed, t2t_feed, act_ret_to_t2t)

    # filtered = [op for op in t2t_sess.graph.get_operations() if '/encoder/layer_0/self_attention' in op.name and op.type == 'MatMul']
    # filtered = [op for op in rnn.engine.tf_session.graph.get_operations() if 'enc_1_self_att_/' in op.name and op.type == 'MatMul']
    # for op in filtered: print(op.name)
    #

    ipdb.set_trace()


def eval_ret_tensor(ret_lt_name, ret_feed):
    ret_act = rnn.engine.tf_session.graph.get_tensor_by_name(ret_lt_name)
    return rnn.engine.tf_session.run(ret_act, ret_feed)


def compare_acts(network, t2t_sess, ret_feed, t2t_feed, act_ret_to_t2t):
    for ret_lt_name, t2t_t_names in act_ret_to_t2t.items():
        ######################################################
        print(ret_lt_name, ":")
        transp = False
        if (
            ":" in ret_lt_name
        ):  # activations are either extracted from graph (more fine-grained) or from a returnn layer
            ret_act = rnn.engine.tf_session.graph.get_tensor_by_name(ret_lt_name)
            transp = True
        else:
            ret_act = network.layers[ret_lt_name].output.get_placeholder_as_batch_major()
        ret_np = rnn.engine.tf_session.run(ret_act, ret_feed)
        print(ret_np.shape)
        if len(ret_np.shape) < 3:
            print(ret_np)
        else:
            if transp:
                ret_np = ret_np.transpose(1, 0, 2)
                print(ret_np.shape)
            print(ret_np[:, :, 0:12])

        #####################################################
        print(t2t_t_names, ":")
        if isinstance(t2t_t_names, tuple):
            t2t_np = numpy.concatenate(
                [t2t_sess.run(t2t_sess.graph.get_tensor_by_name(t2t_t_name), t2t_feed) for t2t_t_name in t2t_t_names],
                axis=1,
            )
        else:
            t2t_np = t2t_sess.run(t2t_sess.graph.get_tensor_by_name(t2t_t_names), t2t_feed)
        print(t2t_np.shape)
        if len(ret_np.shape) < 3:
            print(t2t_np)
        else:
            print(t2t_np[:, :, 0:12])

        print(
            "allclose:",
            numpy.allclose(
                ret_np,
                t2t_np,
                rtol=1.0e-3,
                atol=1.0e-6,
            ),
        )

        if "enc_1_self_att_/dot/MatMul:0" in ret_lt_name:
            print("Calculating attention for first head manually:")
            q = t2t_np[:, 0:32]
            q = q * ((256 / 8) ** (-0.5))
            k = t2t_np[:, 256 * 1 + 0 : 256 * 1 + 32]
            energy = numpy.dot(q, k.T)
            print(energy)
            print("after softmax")
            print(rnn.engine.tf_session.run(tf.nn.softmax(energy, axis=-1)))

        print(
            "-----------------------------------------------------------------------------------------------------------"
        )


act_ret_to_t2t = {
    "enc_1_self_att_laynorm": "transformer/parallel_0_4/transformer/transformer/body/encoder/layer_0/self_attention/layer_prepostprocess/layer_norm/add_1:0",
    "enc_1_self_att_out": "transformer/parallel_0_4/transformer/transformer/body/encoder/layer_0/self_attention/layer_postprocess/add:0",
    "enc_N_self_att_laynorm": "transformer/parallel_0_4/transformer/transformer/body/encoder/layer_1/self_attention/layer_prepostprocess/layer_norm/add_1:0",
    "enc_N_self_att_out": "transformer/parallel_0_4/transformer/transformer/body/encoder/layer_1/self_attention/layer_postprocess/add:0",
    "enc_1_self_att_/Softmax:0": "transformer/parallel_0_4/transformer/transformer/body/encoder/layer_0/self_attention/multihead_attention/dot_product_attention/Softmax:0",
    "encoder": "transformer/parallel_0_4/transformer/transformer/body/encoder/layer_prepostprocess/layer_norm/add_1:0",
    "output/rec/dec_1_self_att_laynorm/add:0": "transformer/parallel_0_4/transformer/transformer/body/decoder/layer_0/self_attention/layer_prepostprocess/layer_norm/add_1:0",
    "output/rec/dec_1_self_att_out/Add:0": "transformer/parallel_0_4/transformer/transformer/body/decoder/layer_0/self_attention/layer_postprocess/add:0",
    "output/rec/dec_N_self_att_out/Add:0": "transformer/parallel_0_4/transformer/transformer/body/decoder/layer_1/self_attention/layer_postprocess/add:0",
    "output/rec/output_prob/linear/dot/MatMul:0": "transformer/parallel_0_4/transformer/transformer/symbol_modality_6115_256_2/softmax/MatMul:0",
}

# act_ret_to_t2t = { 'output/rec/dec_N_self_att_out/Add:0' : 'transformer/parallel_0_4/transformer/transformer/body/decoder/layer_1/self_attention/layer_postprocess/add:0', }
# act_ret_to_t2t = {'enc_1_self_att_/dot/MatMul:0' : tuple ('transformer/parallel_0_4/transformer/transformer/body/encoder/layer_0/self_attention/multihead_attention/%s/Tensordot/MatMul:0' % t for t in ['q', 'k', 'v']),}


# maps names of trainable params (returnn to t2t)
ret_to_t2t = {
    "output/rec/target_embed_raw/W": "transformer/symbol_modality_6115_256/target_emb/weights_0",
    "source_embed_raw/W": "transformer/symbol_modality_6115_256/input_emb/weights_0",
    "encoder/scale": "transformer/body/encoder/layer_prepostprocess/layer_norm/layer_norm_scale",
    "encoder/bias": "transformer/body/encoder/layer_prepostprocess/layer_norm/layer_norm_bias",
    "output/rec/decoder/scale": "transformer/body/decoder/layer_prepostprocess/layer_norm/layer_norm_scale",
    "output/rec/decoder/bias": "transformer/body/decoder/layer_prepostprocess/layer_norm/layer_norm_bias",
    "output/rec/output_prob/W": "transformer/symbol_modality_6115_256/softmax/weights_0",  # has to be transposed
    "enc_1_self_att_laynorm/scale": "transformer/body/encoder/layer_0/self_attention/layer_prepostprocess/layer_norm/layer_norm_scale",
    "enc_1_self_att_laynorm/bias": "transformer/body/encoder/layer_0/self_attention/layer_prepostprocess/layer_norm/layer_norm_bias",
    "enc_1_self_att_/QKV": (
        "transformer/body/encoder/layer_0/self_attention/multihead_attention/q/kernel",
        "transformer/body/encoder/layer_0/self_attention/multihead_attention/k/kernel",
        "transformer/body/encoder/layer_0/self_attention/multihead_attention/v/kernel",
    ),
    "enc_1_self_att_lin/W": "transformer/body/encoder/layer_0/self_attention/multihead_attention/output_transform/kernel",
    "enc_1_ff_laynorm/scale": "transformer/body/encoder/layer_0/ffn/layer_prepostprocess/layer_norm/layer_norm_scale",
    "enc_1_ff_laynorm/bias": "transformer/body/encoder/layer_0/ffn/layer_prepostprocess/layer_norm/layer_norm_bias",
    "enc_1_ff_conv1/W": "transformer/body/encoder/layer_0/ffn/conv1/kernel",
    "enc_1_ff_conv1/b": "transformer/body/encoder/layer_0/ffn/conv1/bias",
    "enc_1_ff_conv2/W": "transformer/body/encoder/layer_0/ffn/conv2/kernel",
    "enc_1_ff_conv2/b": "transformer/body/encoder/layer_0/ffn/conv2/bias",
    "enc_N_self_att_laynorm/scale": "transformer/body/encoder/layer_1/self_attention/layer_prepostprocess/layer_norm/layer_norm_scale",
    "enc_N_self_att_laynorm/bias": "transformer/body/encoder/layer_1/self_attention/layer_prepostprocess/layer_norm/layer_norm_bias",
    "enc_N_self_att_/QKV": (
        "transformer/body/encoder/layer_1/self_attention/multihead_attention/q/kernel",
        "transformer/body/encoder/layer_1/self_attention/multihead_attention/k/kernel",
        "transformer/body/encoder/layer_1/self_attention/multihead_attention/v/kernel",
    ),
    "enc_N_self_att_lin/W": "transformer/body/encoder/layer_1/self_attention/multihead_attention/output_transform/kernel",
    "enc_N_ff_laynorm/scale": "transformer/body/encoder/layer_1/ffn/layer_prepostprocess/layer_norm/layer_norm_scale",
    "enc_N_ff_laynorm/bias": "transformer/body/encoder/layer_1/ffn/layer_prepostprocess/layer_norm/layer_norm_bias",
    "enc_N_ff_conv1/W": "transformer/body/encoder/layer_1/ffn/conv1/kernel",
    "enc_N_ff_conv1/b": "transformer/body/encoder/layer_1/ffn/conv1/bias",
    "enc_N_ff_conv2/W": "transformer/body/encoder/layer_1/ffn/conv2/kernel",
    "enc_N_ff_conv2/b": "transformer/body/encoder/layer_1/ffn/conv2/bias",
    "output/rec/dec_1_self_att_laynorm/scale": "transformer/body/decoder/layer_0/self_attention/layer_prepostprocess/layer_norm/layer_norm_scale",
    "output/rec/dec_1_self_att_laynorm/bias": "transformer/body/decoder/layer_0/self_attention/layer_prepostprocess/layer_norm/layer_norm_bias",
    "output/rec/dec_1_self_att_/QKV": (
        "transformer/body/decoder/layer_0/self_attention/multihead_attention/q/kernel",
        "transformer/body/decoder/layer_0/self_attention/multihead_attention/k/kernel",
        "transformer/body/decoder/layer_0/self_attention/multihead_attention/v/kernel",
    ),
    "output/rec/dec_1_self_att_lin/W": "transformer/body/decoder/layer_0/self_attention/multihead_attention/output_transform/kernel",
    "output/rec/dec_1_att_laynorm/scale": "transformer/body/decoder/layer_0/encdec_attention/layer_prepostprocess/layer_norm/layer_norm_scale",
    "output/rec/dec_1_att_laynorm/bias": "transformer/body/decoder/layer_0/encdec_attention/layer_prepostprocess/layer_norm/layer_norm_bias",
    "output/rec/dec_1_att_query0/W": "transformer/body/decoder/layer_0/encdec_attention/multihead_attention/q/kernel",
    "dec_1_att_key0/W": "transformer/body/decoder/layer_0/encdec_attention/multihead_attention/k/kernel",
    "dec_1_att_value0/W": "transformer/body/decoder/layer_0/encdec_attention/multihead_attention/v/kernel",
    "output/rec/dec_1_att_lin/W": "transformer/body/decoder/layer_0/encdec_attention/multihead_attention/output_transform/kernel",
    "output/rec/dec_1_ff_laynorm/scale": "transformer/body/decoder/layer_0/ffn/layer_prepostprocess/layer_norm/layer_norm_scale",
    "output/rec/dec_1_ff_laynorm/bias": "transformer/body/decoder/layer_0/ffn/layer_prepostprocess/layer_norm/layer_norm_bias",
    "output/rec/dec_1_ff_conv1/W": "transformer/body/decoder/layer_0/ffn/conv1/kernel",
    "output/rec/dec_1_ff_conv1/b": "transformer/body/decoder/layer_0/ffn/conv1/bias",
    "output/rec/dec_1_ff_conv2/W": "transformer/body/decoder/layer_0/ffn/conv2/kernel",
    "output/rec/dec_1_ff_conv2/b": "transformer/body/decoder/layer_0/ffn/conv2/bias",
    "output/rec/dec_N_self_att_laynorm/scale": "transformer/body/decoder/layer_1/self_attention/layer_prepostprocess/layer_norm/layer_norm_scale",
    "output/rec/dec_N_self_att_laynorm/bias": "transformer/body/decoder/layer_1/self_attention/layer_prepostprocess/layer_norm/layer_norm_bias",
    "output/rec/dec_N_self_att_/QKV": (
        "transformer/body/decoder/layer_1/self_attention/multihead_attention/q/kernel",
        "transformer/body/decoder/layer_1/self_attention/multihead_attention/k/kernel",
        "transformer/body/decoder/layer_1/self_attention/multihead_attention/v/kernel",
    ),
    "output/rec/dec_N_self_att_lin/W": "transformer/body/decoder/layer_1/self_attention/multihead_attention/output_transform/kernel",
    "output/rec/dec_N_att_laynorm/scale": "transformer/body/decoder/layer_1/encdec_attention/layer_prepostprocess/layer_norm/layer_norm_scale",
    "output/rec/dec_N_att_laynorm/bias": "transformer/body/decoder/layer_1/encdec_attention/layer_prepostprocess/layer_norm/layer_norm_bias",
    "output/rec/dec_N_att_query0/W": "transformer/body/decoder/layer_1/encdec_attention/multihead_attention/q/kernel",
    "dec_N_att_key0/W": "transformer/body/decoder/layer_1/encdec_attention/multihead_attention/k/kernel",
    "dec_N_att_value0/W": "transformer/body/decoder/layer_1/encdec_attention/multihead_attention/v/kernel",
    "output/rec/dec_N_att_lin/W": "transformer/body/decoder/layer_1/encdec_attention/multihead_attention/output_transform/kernel",
    "output/rec/dec_N_ff_laynorm/scale": "transformer/body/decoder/layer_1/ffn/layer_prepostprocess/layer_norm/layer_norm_scale",
    "output/rec/dec_N_ff_laynorm/bias": "transformer/body/decoder/layer_1/ffn/layer_prepostprocess/layer_norm/layer_norm_bias",
    "output/rec/dec_N_ff_conv1/W": "transformer/body/decoder/layer_1/ffn/conv1/kernel",
    "output/rec/dec_N_ff_conv1/b": "transformer/body/decoder/layer_1/ffn/conv1/bias",
    "output/rec/dec_N_ff_conv2/W": "transformer/body/decoder/layer_1/ffn/conv2/kernel",
    "output/rec/dec_N_ff_conv2/b": "transformer/body/decoder/layer_1/ffn/conv2/bias",
}


if __name__ == "__main__":
    better_exchook.install()
    main()



"""
This package contains modules which all provide different :class:`Dataset` implementations.
A :class:`Dataset` is totally independent from the backend engine used (TF or Theano),
and also mostly independent from other parts of RETURNN, except some utils and logging.
"""

# Make available the most basic API directly here.
from .basic import Dataset, init_dataset, init_dataset_via_str



"""
Hooks logic
"""

from __future__ import annotations
from typing import Optional, Any, Dict, Callable, TypeVar, Tuple
from types import MethodType
from collections import OrderedDict
import weakref


__all__ = ["setup_post_hook_on_method", "RemovableHandle"]


T = TypeVar("T")


def setup_post_hook_on_method(
    obj: T,
    attr: str,
    hook: Callable[[T, Tuple[Any, ...], Dict[str, Any], Any], Optional[Any]],
    *,
    prepend: bool = False,
) -> RemovableHandle:
    """
    :param obj: e.g. module
    :param attr: attribute name to get the method, via ``getattr(obj, attr)``
    :param hook:
    :param prepend: if there are multiple hooks, this will be registered in front of all, otherwise at the end
    """
    method = MethodWithHooks.get(obj, attr)
    return _register_new_hook(method, method.post_hooks, hook, prepend=prepend)


class MethodWithHooks:
    """
    Method with hooks

    (Not sure yet whether this should be exposed in the API...)
    """

    @classmethod
    def get(cls, obj: T, attr: str) -> MethodWithHooks:
        """get existing or init new :class:`MethodWithHooks`"""
        method = getattr(obj, attr)
        if not isinstance(method, MethodWithHooks):
            assert (
                isinstance(method, MethodType)
                and method.__self__ is obj
                and method.__func__ is getattr(obj.__class__, attr)
            ), (
                f"{obj}.{attr} is {method}, {method.__self__} is not obj"
                f" or {method.__func__} is not {getattr(obj.__class__, attr)}"
            )  # not necessary, just what we expect here...
            method = MethodWithHooks(obj, attr)
            method.setup()
        return method

    def __init__(self, obj: T, attr: str):
        """
        :param obj:
        :param attr:
        """
        self.obj = obj
        self.attr = attr
        self.post_hooks: Dict[int, Callable[[T, Tuple[Any, ...], Dict[str, Any], Any], Optional[Any]]] = OrderedDict()

    def setup(self):
        """setup on object"""
        setattr(self.obj, self.attr, self)
        if self.attr == "__call__":
            self.obj.__class__.__call__ = _CallWrapperClass(self.obj.__class__)

    def uninstall(self):
        """undo the setup"""
        assert not self.post_hooks
        delattr(self.obj, self.attr)
        if self.attr == "__call__":
            call_wrapper = self.obj.__class__.__call__
            assert isinstance(call_wrapper, _CallWrapperClass)
            self.obj.__class__.__call__ = call_wrapper.orig_call

    def has_any_hook(self) -> bool:
        """whether there is any hook"""
        return bool(self.post_hooks)

    def __call__(self, *args, **kwargs):
        function = getattr(self.obj.__class__, self.attr)
        if isinstance(function, _CallWrapperClass):
            function = function.orig_call
        result = function(self.obj, *args, **kwargs)
        for hook in self.post_hooks.values():
            result = hook(self.obj, args, kwargs, result)
        return result


class _CallWrapperClass:
    """
    We need special handling for ``obj(*args, **kwargs)`` - simply wrapping ``__call__`` is not enough.
    An instance of this class here would overwrite the obj.__class__.__call__.
    """

    def __init__(self, cls):
        self.cls = cls
        self.orig_call = getattr(cls, "__call__")

    def __call__(self, obj, *args, **kwargs):  # called on the class
        call_func = getattr(obj, "__call__")
        if isinstance(call_func, MethodWithHooks):  # obj has hooks installed
            return call_func(*args, **kwargs)
        # This is what we would expect currently, but the user might have dome some weird things,
        # e.g. overwritten obj.__call__ or obj.__class__.__call__ in some other way,
        # so maybe extend this then, if we encounter this.
        # (Only do it when we really now the case properly, and add a test case for it.)
        assert isinstance(call_func, MethodType)
        assert call_func.__func__ is self
        assert call_func.__self__ is obj
        return self.orig_call(obj, *args, **kwargs)

    def __get__(self, obj, objtype=None):
        if obj is not None:
            return MethodType(self, obj)
        return self


def _register_new_hook(
    method: MethodWithHooks,
    hooks_dict: Dict[int, Any],
    hook: Callable[[T, Tuple[Any, ...], Dict[str, Any], Any], Optional[Any]],
    *,
    prepend: bool = False,
) -> RemovableHandle:
    """register"""
    handle = RemovableHandle(method, hooks_dict)
    hooks_dict[handle.id] = hook
    if prepend:
        assert isinstance(hooks_dict, OrderedDict)
        hooks_dict.move_to_end(handle.id, last=False)
    return handle


class RemovableHandle:
    """
    A handle which provides the capability to remove a hook.

    Code partly copied from PyTorch.
    """

    id: int
    next_id: int = 0

    def __init__(self, method: MethodWithHooks, hooks_dict: Dict[int, Any]):
        self.method_ref = weakref.ref(method)
        self.hooks_dict_ref = weakref.ref(hooks_dict)
        self.id = RemovableHandle.next_id
        RemovableHandle.next_id += 1

    def remove(self) -> None:
        """remove hook"""
        method: MethodWithHooks = self.method_ref()
        hooks_dict = self.hooks_dict_ref()
        if hooks_dict is not None and self.id in hooks_dict:
            del hooks_dict[self.id]
            if method and not method.has_any_hook():  # we removed the last hook
                method.uninstall()



#!/usr/bin/env python3

"""
Scale tuning
"""

from __future__ import annotations
import argparse
import os
import numpy as np
from dataclasses import dataclass
from typing import Optional, Union, Dict, List, Tuple, Set

import _setup_returnn_env  # noqa

from returnn.config import set_global_config, Config
from returnn.log import log
from returnn.util.basic import describe_returnn_version, BehaviorVersion
from returnn.util import better_exchook
from returnn.tensor import Tensor, Dim
import returnn.frontend as rf
import torch
import json


def main():
    """main"""
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument("--names", required=True, nargs="+", help="names for scores")
    arg_parser.add_argument("--scores", required=True, nargs="+", help="score-paths")
    arg_parser.add_argument("--evaluation", required=True, help="'edit_distance'")
    arg_parser.add_argument("--ref", help="ref-path")
    arg_parser.add_argument("--fixed-scales", nargs="*", help="(name, scale) pairs, fixed scales")
    arg_parser.add_argument("--negative-scales", nargs="*", help="list of names, negative scales")
    arg_parser.add_argument("--scale-relative-to", nargs="*", help="(name, other) pairs, scale relative to")
    arg_parser.add_argument("--max-scale", type=float, default=2.0)
    arg_parser.add_argument("--min-scales", nargs="*", help="(name, scale) pairs, min scales")
    arg_parser.add_argument("--max-scales", nargs="*", help="(name, scale) pairs, max scales")
    arg_parser.add_argument("--num-iterations", type=int, default=10)
    arg_parser.add_argument("--num-steps", type=int, default=21)
    arg_parser.add_argument("--search-mode", default="auto", help="'auto', 'grid', 'single'")
    arg_parser.add_argument("--device", default="cpu", help="auto, cpu, cuda, ...")
    arg_parser.add_argument("--batch-size", type=int, default=1024)
    arg_parser.add_argument("--output-scales", help="file to write (relative) scales into")
    arg_parser.add_argument("--output-real-scales", help="file to write real scales into")
    arg_parser.add_argument("--output-grid-plot", help="file to write grid plot into")
    args = arg_parser.parse_args()

    print(f"{os.path.basename(__file__)}, RETURNN {describe_returnn_version()}")

    config = Config()
    set_global_config(config)

    log.initialize(verbosity=[5])
    BehaviorVersion.set_min_behavior_version(22)
    better_exchook.install()
    rf.select_backend_torch()

    try:
        # noinspection PyUnresolvedReferences,PyPackageRequirements
        import lovely_tensors

        lovely_tensors.monkey_patch()
    except ImportError:
        pass  # ignore

    names = list(args.names)
    names_dim = Dim(len(names), name="names")

    # Resort scale names, such that we first have the fixed scales, then the scale_relative_to.
    fixed_scales_names = set()
    if args.fixed_scales:
        assert len(args.fixed_scales) % 2 == 0
        for name, scale in zip(args.fixed_scales[::2], args.fixed_scales[1::2]):
            fixed_scales_names.add(name)
    scale_relative_to_target_names = set()
    if args.scale_relative_to:
        assert len(args.scale_relative_to) % 2 == 0
        for name, other in zip(args.scale_relative_to[::2], args.scale_relative_to[1::2]):
            scale_relative_to_target_names.add(other)
    names.sort(key=lambda key: (key not in fixed_scales_names, key not in scale_relative_to_target_names))
    print("Names:", names)

    # Prepare scale relation info, based on scale/name idx.
    fixed_scales: Dict[int, float] = {}
    if args.fixed_scales:
        assert len(args.fixed_scales) % 2 == 0
        for name, scale in zip(args.fixed_scales[::2], args.fixed_scales[1::2]):
            name_idx = names.index(name)
            assert name_idx not in fixed_scales
            fixed_scales[name_idx] = float(scale)
    scale_relative_to: Dict[int, int] = {}  # name idx -> other name idx (before)
    if args.scale_relative_to:
        assert len(args.scale_relative_to) % 2 == 0
        for name, other in zip(args.scale_relative_to[::2], args.scale_relative_to[1::2]):
            name_idx = names.index(name)
            other_idx = names.index(other)
            assert name_idx not in scale_relative_to
            assert name_idx not in fixed_scales
            assert other_idx not in fixed_scales
            assert other_idx < name_idx, "make sure, when --scale-relative-to, that the other name is before"
            scale_relative_to[name_idx] = other_idx
    neg_scales: Set[int] = set()
    if args.negative_scales:
        for name in args.negative_scales:
            name_idx = names.index(name)
            assert name_idx not in fixed_scales
            neg_scales.add(name_idx)

    print("Num scales to search over:", len(names) - len(fixed_scales))
    search_mode = args.search_mode
    if search_mode == "auto":
        if len(names) - len(fixed_scales) == 2:
            search_mode = "grid"
        else:
            search_mode = "single"
    print("Search mode:", search_mode)

    # Load data
    vocab: Dict[str, int] = {}
    hyps: Dict[str, List[List[int]]] = {}  # seq_tag -> beam of seqs
    scores: Dict[str, Dict[str, List[float]]] = {}  # name -> seq_tag -> beam of score
    assert args.scores and len(args.scores) == len(names)
    for name, path in zip(args.names, args.scores):
        if not os.path.exists(path):
            raise FileNotFoundError(f"score file not found: {path}, for name: {name}")
        scores[name], hyps = _load_text_dict_hyps_file(
            path, name=name, vocab=vocab, expect_same_hyps=hyps if hyps else None
        )
        assert scores[name] and hyps
    print("num seqs:", len(hyps))

    assert scores and hyps
    print("len vocab after reading hyps:", len(vocab))
    assert vocab, "no labels found?"
    ref: Optional[Dict[str, List[int]]] = (
        _load_text_dict_file(args.ref, name="ref", vocab=vocab) if args.ref is not None else None
    )
    print("len vocab after reading ref:", len(vocab))
    if len(vocab) < 2**15:
        dtype = torch.int16
    elif len(vocab) < 2**31:
        dtype = torch.int32
    elif len(vocab) < 2**63:
        dtype = torch.int64
    else:
        raise ValueError(f"vocab too large: {len(vocab)}")
    print("dtype:", dtype)

    if ref:
        total_ref_seq_len = sum(len(seq) for seq in ref.values())
        print("total ref seq len:", total_ref_seq_len)
        avg_ref_seq_len = total_ref_seq_len / len(ref)
        print("avg ref seq len:", avg_ref_seq_len)
    else:
        avg_ref_seq_len = None

    # Sort by lengths (reversed) to be able to prepare batches without too much padding.
    if ref:
        seq_list_ordered_by_len = sorted(ref.keys(), key=lambda tag: len(ref[tag]), reverse=True)
    else:
        seq_list_ordered_by_len = sorted(hyps.keys(), key=lambda tag: len(hyps[tag][0]), reverse=True)

    if args.device == "auto":
        dev_s = "cuda" if torch.cuda.is_available() else "cpu"
    else:
        dev_s = args.device
    dev = torch.device(dev_s)
    print("Device:", dev)
    _report_dev_memory(dev)

    # Prepare batches, load all into device memory
    # Note: If we get GPU OOM here, we could instead keep this on CPU... Currently, not yet implemented...
    print("Preparing batches, calculating evaluation on all hyps...")
    batches = []
    for i in range(0, len(seq_list_ordered_by_len), args.batch_size):
        print(
            f"Batch {len(batches)}, seqs {i} - {min(i + args.batch_size, len(seq_list_ordered_by_len))}"
            f" / {len(seq_list_ordered_by_len)}, {i / len(seq_list_ordered_by_len) * 100:.1f}%"
        )
        batch_seq_tag_list: List[str] = seq_list_ordered_by_len[i : i + args.batch_size]

        beam_sizes_ = [len(hyps[tag]) for tag in batch_seq_tag_list]
        beam_sizes_t = torch.tensor(beam_sizes_, dtype=torch.int32)  # [Batch], int32
        hyps_seq_lens_: List[List[int]] = [
            [len(hyps[tag][beam]) for beam in range(len(hyps[tag]))] for tag in batch_seq_tag_list
        ]
        hyps_seq_lens_t = _make_padded_tensor_2d(hyps_seq_lens_, dtype=torch.int32)  # [Batch,Beam], int32
        hyps__: List[List[List[int]]] = [hyps[tag] for tag in batch_seq_tag_list]
        hyps_t = _make_padded_tensor_3d(hyps__, dtype=dtype, device=dev)  # [Batch,Beam,HypSeq], int16|int32|int64

        batch_scores: List[List[List[float]]] = [[scores[name][tag] for tag in batch_seq_tag_list] for name in names]
        batch_scores_t = _make_padded_tensor_3d(
            batch_scores, dtype=torch.float32, device=dev
        )  # [Names,Batch,Beam], float32

        if ref is not None:
            ref_: List[List[int]] = [ref[tag] for tag in batch_seq_tag_list]
            ref_t = _make_padded_tensor_2d(ref_, dtype=dtype, device=dev)  # [Batch,RefSeq], int16|int32|int64
            ref_seq_lens = [len(ref[tag]) for tag in batch_seq_tag_list]
            ref_seq_lens_t = torch.tensor(ref_seq_lens, dtype=torch.int32)  # [Batch], int32
        else:
            ref_t = None
            ref_seq_lens_t = None

        batch_dim = Dim(len(batch_seq_tag_list), name="batch")
        beam_sizes_rf = rf.convert_to_tensor(beam_sizes_t, dims=[batch_dim])
        beam_dim = Dim(beam_sizes_rf, name="beam")
        hyps_seq_lens_rf = rf.convert_to_tensor(hyps_seq_lens_t, dims=[batch_dim, beam_dim])
        hyps_seq_dim = Dim(hyps_seq_lens_rf, name="hyps_seq")
        hyps_rf = rf.convert_to_tensor(hyps_t, dims=[batch_dim, beam_dim, hyps_seq_dim])
        if ref_t is not None:
            ref_seq_lens_rf = rf.convert_to_tensor(ref_seq_lens_t, dims=[batch_dim])
            ref_seq_dim = Dim(ref_seq_lens_rf, name="ref_seq")
            ref_rf = rf.convert_to_tensor(ref_t, dims=[batch_dim, ref_seq_dim])
        else:
            ref_seq_dim = None
            ref_rf = None

        if args.evaluation == "edit_distance":
            assert ref_rf is not None, "need --ref for edit_distance"
            res = rf.edit_distance(ref_rf, ref_seq_dim, hyps_rf, hyps_seq_dim)
            assert res.dims_set == {batch_dim, beam_dim}
            res = res.copy_transpose((batch_dim, beam_dim))
            res = rf.cast(res, "float64")
            # Using avg_ref_seq_len makes this numerically more stable, as the numbers are not so small,
            # and we can use reduce_mean below, which is also more stable.
            res /= avg_ref_seq_len  # WER / avg_len = (WER / total_len) * num_seqs -> can use reduce_mean
        else:
            raise ValueError(f"unknown evaluation {args.evaluation!r}")

        batch_scores_rf = rf.convert_to_tensor(batch_scores_t, dims=[names_dim, batch_dim, beam_dim])

        batches.append(
            Batch(
                seq_tags=batch_seq_tag_list,
                batch_dim=batch_dim,
                beam_dim=beam_dim,
                hyps_eval=res,
                names_dim=names_dim,
                scores=batch_scores_rf,
            )
        )
    print("num batches:", len(batches))
    _report_dev_memory(dev)

    def _eval_max_min(*, use_max: bool) -> float:
        hyps_eval_ts = []
        for batch in batches:
            hyps_eval = (rf.reduce_max if use_max else rf.reduce_min)(
                batch.hyps_eval, axis=batch.beam_dim
            )  # [Batch], float64
            assert hyps_eval.dims_set == {batch.batch_dim}
            hyps_eval_ts.append(hyps_eval.raw_tensor)
        hyps_eval_t = torch.concatenate(hyps_eval_ts)  # [NumSeqs], float64
        hyps_eval_t = torch.mean(hyps_eval_t)  # scalar, float64
        return hyps_eval_t.cpu().item()

    best_eval, worst_eval = _eval_max_min(use_max=False), _eval_max_min(use_max=True)
    print(f"Best {args.evaluation}: {best_eval}, worst {args.evaluation}: {worst_eval}")

    scales_min = [0.0] * len(names)
    scales_max = [args.max_scale] * len(names)
    if args.min_scales:
        assert len(args.min_scales) % 2 == 0
        for name, scale in zip(args.min_scales[::2], args.min_scales[1::2]):
            scales_min[names.index(name)] = float(scale)
    if args.max_scales:
        assert len(args.max_scales) % 2 == 0
        for name, scale in zip(args.max_scales[::2], args.max_scales[1::2]):
            scales_max[names.index(name)] = float(scale)

    scales = [(scales_min[i] + scales_max[i]) / 2 for i in range(len(names))]
    for name_idx, scale in fixed_scales.items():
        scales[name_idx] = scale

    def _real_scales(scales_: List[float]):
        scales_ = list(scales_)
        for i_ in range(len(scales_)):
            if i_ in scale_relative_to:
                scales_[i_] *= scales_[scale_relative_to[i_]]
            if i_ in neg_scales:
                scales_[i_] *= -1
        return scales_

    def _eval_for_scales(scales_: List[float]) -> float:
        real_scales_t = torch.tensor(_real_scales(scales_), dtype=torch.float32).to(dev)  # [Names], float32
        real_scales = rf.convert_to_tensor(real_scales_t, dims=[names_dim])
        hyps_eval_ts = []
        for batch in batches:
            scores_scaled = rf.matmul(batch.scores, real_scales, reduce=names_dim)  # [Batch,Beam], float32
            beam_idx = rf.reduce_argmax(scores_scaled, axis=batch.beam_dim)  # [Batch] -> Beam, int32
            hyps_eval = rf.gather(batch.hyps_eval, indices=beam_idx)  # [Batch], float64
            assert hyps_eval.dims_set == {batch.batch_dim}
            hyps_eval_ts.append(hyps_eval.raw_tensor)
        hyps_eval_t = torch.concatenate(hyps_eval_ts)  # [NumSeqs], float64
        hyps_eval_t = torch.mean(hyps_eval_t)  # scalar, float64
        return hyps_eval_t.cpu().item()

    if search_mode == "single":
        print("Search mode single, each dimension/scale optimized separately")
        for iter_idx in range(args.num_iterations):
            print("*** Iter", iter_idx)
            has_change = False
            for scale_idx in range(len(scales) - 1, -1, -1):
                if scale_idx in fixed_scales:
                    continue
                print(f"** Scale {names[scale_idx]} in range {scales_min[scale_idx]} - {scales_max[scale_idx]}")
                print(
                    "  Other scales:",
                    ", ".join(f"{names[i]}: {scales[i]}" for i in range(len(scales)) if i != scale_idx),
                )
                evals: List[Tuple[float, float]] = []  # (eval, scale)
                for scale in np.linspace(scales_min[scale_idx], scales_max[scale_idx], num=args.num_steps):
                    scales[scale_idx] = scale
                    eval_ = _eval_for_scales(scales)
                    print(f"Scale {names[scale_idx]}: {scale}, {args.evaluation}: {eval_}")
                    evals.append((eval_, scale))
                eval_p = np.percentile([eval_ for eval_, _ in evals], 5)
                print("Eval p5:", eval_p, "best:", min(eval_ for eval_, _ in evals))
                prev_min, prev_max = scales_min[scale_idx], scales_max[scale_idx]
                scales_min[scale_idx] = min(scale for eval_, scale in evals if eval_ <= eval_p * 1.0001)
                scales_max[scale_idx] = max(scale for eval_, scale in evals if eval_ <= eval_p * 1.0001)
                print(f"New {names[scale_idx]} scales min/max:", scales_min[scale_idx], scales_max[scale_idx])
                scales[scale_idx] = (scales_min[scale_idx] + scales_max[scale_idx]) / 2
                if prev_min != scales_min[scale_idx] or prev_max != scales_max[scale_idx]:
                    has_change = True
                else:
                    print(f"No change for scale {names[scale_idx]}")
            if not has_change:
                print(f"No change in this iteration {iter_idx}, stop")
                break
    elif search_mode == "grid":
        print("Search mode grid, all dimensions/scales optimized together")
        for iter_idx in range(args.num_iterations):
            print("*** Iter", iter_idx)
            has_change = False
            scale_indices = []
            spaces = []
            for scale_idx in range(len(names)):
                if scale_idx in fixed_scales:
                    continue
                scale_indices.append(scale_idx)
                spaces.append(np.linspace(scales_min[scale_idx], scales_max[scale_idx], num=args.num_steps))
            evals: List[Tuple[float, List[float]]] = []  # (eval, scales)
            best_eval_so_far = np.inf
            spaces = np.meshgrid(*spaces)
            it = np.nditer(spaces)
            with it:
                for scale_values in it:
                    assert len(scale_values) == len(scale_indices)
                    scales = [0.0] * len(names)
                    for scale_idx, scale in fixed_scales.items():
                        scales[scale_idx] = scale
                    for scale_idx, scale in zip(scale_indices, scale_values):
                        scales[scale_idx] = float(scale)
                    eval_ = _eval_for_scales(scales)
                    if eval_ < best_eval_so_far:
                        best_eval_so_far = eval_
                        print(f"New best {args.evaluation}: {eval_}, scales: {scales}")
                    evals.append((eval_, scales))
            if args.output_grid_plot:
                assert len(scale_indices) == 2, "only implemented for 2 scales"
                _plot_grid(
                    evals,
                    scale_indices=scale_indices,
                    title="",
                    cbar_label=args.evaluation,
                    x_axis_name=names[scale_indices[0]],
                    y_axis_name=names[scale_indices[1]],
                    out_plot_filename=f"{args.output_grid_plot}.{iter_idx}.pdf",
                )
            eval_p = np.percentile([eval_ for eval_, _ in evals], 5)
            print(
                f"Evaluated grid size {len(evals)}, eval p5: {eval_p},",
                f"best: {min(eval_ for eval_, _ in evals)},",
                f"worst: {max(eval_ for eval_, _ in evals)}",
            )
            for scale_idx in scale_indices:
                prev_min, prev_max = scales_min[scale_idx], scales_max[scale_idx]
                scales_min[scale_idx] = min(scale[scale_idx] for eval_, scale in evals if eval_ <= eval_p * 1.0001)
                scales_max[scale_idx] = max(scale[scale_idx] for eval_, scale in evals if eval_ <= eval_p * 1.0001)
                print(f"New {names[scale_idx]} scales min/max:", scales_min[scale_idx], scales_max[scale_idx])
                if prev_min != scales_min[scale_idx] or prev_max != scales_max[scale_idx]:
                    has_change = True
                else:
                    print(f"No change for scale {names[scale_idx]}")
            # Select current best.
            scales = min(evals)[1]
            if not has_change:
                print(f"No change in this iteration {iter_idx}, stop")
                break
    else:
        raise ValueError(f"unknown search mode {search_mode!r}")

    print("Final scales:")
    for name, scale, real_scale in zip(names, scales, _real_scales(scales)):
        print(f"{name}: {scale} (real: {real_scale})")
    eval_ = _eval_for_scales(scales)
    print(f"Final {args.evaluation}: {eval_}")

    if args.output_scales:
        print("Writing scales to", args.output_scales)
        with open(args.output_scales, "w") as f:
            f.write(json.dumps(dict(zip(names, scales))) + "\n")
    if args.output_real_scales:
        print("Writing real scales to", args.output_real_scales)
        with open(args.output_real_scales, "w") as f:
            f.write(json.dumps(dict(zip(names, _real_scales(scales)))) + "\n")


@dataclass
class Batch:
    """batch"""

    seq_tags: List[str]
    batch_dim: Dim  # scalar, int32
    beam_dim: Dim  # [Batch], int32
    hyps_eval: Tensor  # [Batch,Beam], float64
    names_dim: Dim  # scalar, int32
    scores: Tensor  # [Names,Batch,Beam], float32


def _make_padded_tensor_2d(
    lst: List[List[Union[int, float]]], *, dtype: torch.dtype, device: Optional[torch.device] = None
) -> torch.Tensor:
    max_len = max(len(l_) for l_ in lst)
    res = torch.zeros((len(lst), max_len), dtype=dtype)
    for i, l in enumerate(lst):
        for j, v in enumerate(l):
            res[i, j] = v
    return res.to(device)


def _make_padded_tensor_3d(
    lst: List[List[List[Union[int, float]]]], *, dtype: torch.dtype, device: torch.device
) -> torch.Tensor:
    max_len = max(len(l_) for l_ in lst)
    max_len2 = max(len(l__) for l_ in lst for l__ in l_)
    res = torch.zeros((len(lst), max_len, max_len2), dtype=dtype)
    for i, l in enumerate(lst):
        for j, l2 in enumerate(l):
            for k, v in enumerate(l2):
                res[i, j, k] = v
    return res.to(device)


def _load_text_dict_hyps_file(
    filename: str,
    *,
    name: str,
    vocab: Dict[str, int],
    expect_same_hyps: Optional[Dict[str, List[List[int]]]] = None,
) -> Tuple[Dict[str, List[float]], Dict[str, List[List[int]]]]:
    # See also code in RETURNN TextDictDataset.
    print(f"Loading text dict file {name} from {filename} ...")

    if filename.endswith(".gz"):
        import gzip

        txt = gzip.GzipFile(filename, "rb").read()
    else:
        txt = open(filename, "rb").read()
    data: Dict[str, List[Tuple[float, str]]] = eval(txt, {"inf": float("inf"), "nan": float("nan")})
    assert isinstance(data, dict)
    res_scores = {}
    res_hyps = {}
    if expect_same_hyps is not None:
        assert set(data.keys()) == set(expect_same_hyps.keys())
    for tag, hyps in data.items():
        res_scores[tag], res_hyps[tag] = _hyps_to_indices(
            hyps,
            vocab=vocab,
            expect_same_hyps=expect_same_hyps[tag] if expect_same_hyps else None,
        )
    return res_scores, res_hyps


def _load_text_dict_file(filename: str, *, name: str, vocab: Dict[str, int]) -> Dict[str, List[int]]:
    # See also code in RETURNN TextDictDataset.
    print(f"Loading text dict file {name} from {filename} ...")

    if filename.endswith(".gz"):
        import gzip

        txt = gzip.GzipFile(filename, "rb").read()
    else:
        txt = open(filename, "rb").read()
    data: Dict[str, str] = eval(txt)
    assert isinstance(data, dict)
    res = {}
    for tag, hyps in data.items():
        res[tag] = _hyp_to_indices(hyps, vocab=vocab)
    return res


def _hyps_to_indices(
    hyp: List[Tuple[float, str]],
    *,
    vocab: Dict[str, int],
    expect_same_hyps: Optional[List[List[int]]] = None,
) -> Tuple[List[float], List[List[int]]]:
    assert isinstance(hyp, list)
    if expect_same_hyps is not None:
        assert len(expect_same_hyps) == len(hyp)
    res_scores = []
    res_hyps = []
    for i, (score, hyp_) in enumerate(hyp):
        assert isinstance(score, float) and isinstance(hyp_, str)
        res_scores.append(score)
        seq = _hyp_to_indices(hyp_, vocab=vocab, assert_in_vocab=expect_same_hyps is not None)
        if expect_same_hyps is not None:
            assert seq == expect_same_hyps[i]
        res_hyps.append(seq)
    return res_scores, res_hyps


def _hyp_to_indices(hyp: str, *, vocab: Dict[str, int], assert_in_vocab: bool = False) -> List[int]:
    assert isinstance(hyp, str)
    res = []
    for label in hyp.split():
        if label not in vocab:
            assert not assert_in_vocab, f"unknown label {label}"
            vocab[label] = len(vocab)
        res.append(vocab[label])
    return res


def _report_dev_memory(dev: torch.device):
    import torch
    from returnn.util import basic as rutil

    if dev.type == "cuda":
        stats = [
            f"alloc cur {rutil.human_bytes_size(torch.cuda.memory_allocated(dev))}",
            f"alloc peak {rutil.human_bytes_size(torch.cuda.max_memory_allocated(dev))}",
            f"reserved cur {rutil.human_bytes_size(torch.cuda.memory_reserved(dev))}",
            f"reserved peak {rutil.human_bytes_size(torch.cuda.max_memory_reserved(dev))}",
        ]
        print(f"Memory usage ({dev}):", " ".join(stats))


def _plot_grid(
    evals: List[Tuple[float, List[float]]],
    *,
    scale_indices: List[int],
    title: str,
    cbar_label: str,
    y_axis_name: str,
    x_axis_name: str,
    out_plot_filename: str,
):
    # noinspection PyPackageRequirements
    import matplotlib.pyplot as plt

    # noinspection PyPackageRequirements
    import matplotlib.ticker as ticker

    results = {}  # (x,y) -> z
    for eval_, scales in evals:
        results[tuple([scales[i] for i in scale_indices])] = eval_
    xs = sorted(set(scales[scale_indices[0]] for _, scales in evals))
    ys = sorted(set(scales[scale_indices[1]] for _, scales in evals))

    plt.figure(figsize=(8, 8))

    zs = np.zeros((len(ys), len(xs)))
    for y_idx, y in enumerate(ys):
        for x_idx, x in enumerate(xs):
            zs[y_idx, x_idx] = results[(x, y)]

    best = np.min(zs.flatten())
    worst_limit = best * 1.3

    ax = plt.subplot(1, 1, 1)
    plt.contourf(xs, ys, zs, levels=np.geomspace(best, worst_limit, 30))

    ax.set_title(title)
    ax.set_ylabel(y_axis_name)
    ax.set_xlabel(x_axis_name)
    ax.xaxis.set_major_locator(ticker.AutoLocator())
    ax.xaxis.set_minor_locator(ticker.AutoMinorLocator())
    ax.yaxis.set_major_locator(ticker.AutoLocator())
    ax.yaxis.set_minor_locator(ticker.AutoMinorLocator())

    cbar = plt.colorbar()
    cbar.set_label(cbar_label)

    print("Saving plot to", out_plot_filename)
    plt.savefig(out_plot_filename)


if __name__ == "__main__":
    main()



# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Class to transform an subgraph into another."""

from __future__ import annotations

from copy import deepcopy
from functools import partial
from six import iteritems
from six import string_types
from six import StringIO
from . import reroute
from . import select
from . import subgraph
from . import util
from returnn.tf.util.basic import copy_op
import tensorflow as tf
from tensorflow.python.framework import ops as tf_ops
from tensorflow.python.platform import tf_logging as logging


__all__ = [
    "replace_t_with_placeholder_handler",
    "keep_t_if_possible_handler",
    "assign_renamed_collections_handler",
    "transform_op_if_inside_handler",
    "copy_op_handler",
    "Transformer",
    "TransformerInfo",
    "copy",
    "copy_with_input_replacements",
    "graph_replace",
]


def replace_t_with_placeholder_handler(info, t):
    """Transform a tensor into a placeholder tensor.

    This handler is typically used to transform a subgraph input tensor into a
    placeholder.

    Args:
      info: Transform._TmpInfo instance.
      t: tensor whose input must be transformed into a place holder.
    Returns:
      The tensor generated by the newly created place holder.
    """
    with info.graph_.as_default():
        t_ = util.make_placeholder_from_tensor(t, scope=info.scope_)
    return t_


def keep_t_if_possible_handler(info, t):
    """Transform a tensor into itself (identity) if possible.

    This handler transform a tensor into itself if the source and destination
    graph are the same. Otherwise it will create a placeholder.
    This handler is typically used to transform a hidden input tensors.

    Args:
      info: Transform._TmpInfo instance.
      t: tensor whose input must be transformed into a place holder.
    Returns:
      The tensor generated by the newly created place holder.
    """
    if info.graph is info.graph_:
        return t
    else:
        return replace_t_with_placeholder_handler(info, t)


def assign_renamed_collections_handler(info, elem, elem_):
    """Add the transformed elem to the (renamed) collections of elem.

    A collection is renamed only if is not a known key, as described in
    `tf.compat.v1.GraphKeys`.

    Args:
      info: Transform._TmpInfo instance.
      elem: the original element (`tf.Tensor` or `tf.Operation`)
      elem_: the transformed element
    """
    known_collection_names = util.get_predefined_collection_names()
    for name, collection in iteritems(info.collections):
        if elem not in collection:
            continue

        if name in known_collection_names:
            transformed_name = name
        else:
            transformed_name = info.new_name(name)
        info.graph_.add_to_collection(transformed_name, elem_)


def transform_op_if_inside_handler(info, op, keep_if_possible=True):
    """Transform an optional op only if it is inside the subgraph.

    This handler is typically use to handle original op: it is fine to keep them
    if they are inside the subgraph, otherwise they are just ignored.

    Args:
      info: Transform._TmpInfo instance.
      op: the optional op to transform (or ignore).
      keep_if_possible: re-attach to the original op if possible, that is,
        if the source graph and the destination graph are the same.
    Returns:
      The transformed op or None.
    """
    if op in info.sgv.ops:
        return info.transformed_ops[op]
    else:
        if keep_if_possible and info.graph is info.graph_:
            return op
        else:
            return None


def copy_op_handler(info, op, new_inputs):
    """Copy a `tf.Operation`.

    Args:
      info: Transform._TmpInfo instance.
      op: the `tf.Operation` to be copied.
      new_inputs: The new inputs for this op.

    Returns:
      A `(op, op_outputs)` tuple containing the transformed op and its outputs.
    """
    # The `new_inputs` was added to this function. For compatibility reason,
    # let's raise an error if `new_inputs` is a boolean.
    if isinstance(new_inputs, bool):
        raise TypeError("the `new_inputs` argument must be an iterable.")

    # Transform name:
    name_ = info.new_name(op.name)
    name_ = info.graph_.unique_name(name_)

    op_ = copy_op(op, inputs=new_inputs, graph=info.graph_, name=name_)

    # Original op cannot be finalised here yet. Because some ops require this
    # attribute to exist, we will create a dummy original_op first and then
    # later finalise it with the actual original_op when all the ops have
    # been copied.
    # TODO(fkp): Stop worrying about _original_op and remove this code?
    if op._original_op:
        op_._original_op = op._original_op

    return op_, op_.outputs


class TransformerInfo:
    """ "Contains information about the result of a transform operation."""

    def __init__(self, info):
        """Constructor.

        Args:
          info: an instance of Transformer._TmpInfo containing various internal
            information about the transform operation.
        """
        self._graph = info.graph
        self._scope = info.scope
        self._graph_ = info.graph_
        self._scope_ = info.scope_
        self._transformed_ops = info.transformed_ops
        self._transformed_ts = info.transformed_ts

    def _get_transformed_map(self, top):
        """Return the correct container depending on the type of `top`."""
        if isinstance(top, tf_ops.Operation):
            return self._transformed_ops
        elif isinstance(top, tf.Tensor):
            return self._transformed_ts
        else:
            raise TypeError("Expected a tf.Tensor or a tf.Operation, got a {}".format(type(top)))

    def _transformed_elem(self, original_top, missing_fn=None):
        """Return the transformed op/tensor corresponding to the original one.

        Args:
          original_top: the original tensor/operation.
          missing_fn: function handling the case where the counterpart
            cannot be found. By default, None is returned.
        Returns:
          the transformed tensor/operation (or None if no match is found).
        """
        transformed_map = self._get_transformed_map(original_top)
        if isinstance(original_top, string_types):
            for original, transformed in iteritems(transformed_map):
                if original.name == original_top:
                    return transformed
            return None if missing_fn is None else missing_fn(original_top)
        else:
            if original_top not in transformed_map:
                return None if missing_fn is None else missing_fn(original_top)
            return transformed_map[original_top]

    def _original_elem(self, transformed_top, missing_fn=None):
        """Return the original op/tensor corresponding to the transformed one.

        Args:
          transformed_top: the transformed tensor/operation.
          missing_fn: function handling the case where the counterpart
            cannot be found. By default, None is returned.
        Returns:
          the original tensor/operation (or None if no match is found).
        """
        transformed_map = self._get_transformed_map(transformed_top)
        if isinstance(transformed_top, string_types):
            finder = lambda transformed: transformed.name == transformed_top
        else:
            finder = lambda transformed: transformed == transformed_top
        for original, transformed in iteritems(transformed_map):
            if finder(transformed):
                return original
        return None if missing_fn is None else missing_fn(transformed_top)

    def transformed(self, original, missing_fn=None):
        """Return the transformed op/tensor corresponding to the original one.

        Note that the output of this function mimics the hierarchy
        of its input argument `original`.
        Given an iterable, it returns a list. Given an operation or a tensor,
        it will return an operation or a tensor.

        Args:
          original: the original tensor/operation.
          missing_fn: function handling the case where the counterpart
            cannot be found. By default, None is returned.
        Returns:
          the transformed tensor/operation (or None if no match is found).
        """
        transformed_elem = partial(self._transformed_elem, missing_fn=missing_fn)
        return util.transform_tree(original, transformed_elem)

    def original(self, transformed, missing_fn=None):
        """Return the original op/tensor corresponding to the transformed one.

        Note that the output of this function mimics the hierarchy
        of its input argument `transformed`.
        Given an iterable, it returns a list. Given an operation or a tensor,
        it will return an operation or a tensor.

        Args:
          transformed: the transformed tensor/operation.
          missing_fn: function handling the case where the counterpart
            cannot be found. By default, None is returned.
        Returns:
          the original tensor/operation (or None if no match is found).
        """
        original_elem = partial(self._original_elem, missing_fn=missing_fn)
        return util.transform_tree(transformed, original_elem)

    def __str__(self):
        res = StringIO()
        print("Transform result info:", file=res)
        if self._graph == self._graph_:
            in_place_str = "" if self._scope_ else " IN-PLACE"
            print("  Within graph[{}]{}".format(id(self._graph), in_place_str), file=res)
        else:
            print("  graph[{}] => graph[{}]".format(id(self._graph), id(self._graph_)), file=res)
        if self._scope:
            print("  Relative to source scope: {}".format(self._scope), file=res)
        if self._scope_:
            print("  Scope destination: {}".format(self._scope_), file=res)
        print("Operations mapping:", file=res)
        for op, op_ in iteritems(self._transformed_ops):
            print("  {} => {}".format(op.name, op_.name), file=res)
        return res.getvalue()


class _TmpInfo:
    """Transformer temporary data.

    An instance of this class holds all the information relevant to a call
    to a transformer instance (that is, a call to __call__). An instance
    is created for the life-time of the __call__ function and is passed as
    argument to the handlers.
    """

    def __init__(self, sgv, dst_graph, dst_scope, src_scope):
        self.sgv = sgv
        self.sgv_inputs_set = frozenset(sgv.inputs)
        self.ops = frozenset(sgv.ops)
        self.control_outputs = util.ControlOutputs(sgv.graph)
        self.graph = sgv.graph
        self.scope = src_scope
        self.graph_ = dst_graph
        self.scope_ = dst_scope
        self.transformed_ops = {}
        self.transformed_ts = {}
        self.collections = dict((key, self.graph.get_collection(key)) for key in self.graph.get_all_collection_keys())
        self.cyclic_ops = []
        self.transform_original_op_handler = transform_op_if_inside_handler
        # The graph is transformed op by op, in the same order the original ops
        # were created. However, this is sometimes not possible due to cycles
        # (i.e. while loops). So when the transformer creates a new op whose
        # inputs do not exist yet, temporary placeholders are created and stored
        # in this `tmp_cyclic_ts` container. During a second pass,
        # those temporary tensors are replaced by the proper transformed tensors
        # (see the function `_finalize_cycles`).
        self.tmp_cyclic_ts = []

    def new_name(self, name):
        """Compute a destination name from a source name.

        Args:
          name: the name to be "transformed".
        Returns:
          The transformed name.
        Raises:
          ValueError: if the source scope is used (that is, not an empty string)
            and the source name does not belong to the source scope.
        """
        scope = self.scope
        if not name.startswith(scope):
            raise ValueError("{} does not belong to source scope: {}.".format(name, scope))
        rel_name = name[len(scope) :]
        name_ = self.scope_ + rel_name
        return name_


class Transformer:
    """Transform a subgraph into another one.

    By default, the constructor create a transform which copy a subgraph and
    replaces inputs with placeholders. This behavior can be modified by changing
    the handlers.
    """

    def __init__(self):
        """Transformer constructor.

        The following members can be modified:
        transform_op_handler: handle the transformation of a `tf.Operation`.
          This handler defaults to a simple copy.
        assign_collections_handler: handle the assignment of collections.
          This handler defaults to assigning new collections created under the
          given name-scope.
        transform_external_input_handler: handle the transform of the inputs to
          the given subgraph. This handler defaults to creating placeholders
          instead of the ops just before the input tensors of the subgraph.
        transform_external_hidden_input_handler: handle the transform of the
          hidden inputs of the subgraph, that is, the inputs which are not listed
          in sgv.inputs. This handler defaults to a transform which keep the same
          input if the source and destination graphs are the same, otherwise
          use placeholders.
        transform_original_op_handler: handle the transform of original_op. This
          handler defaults to transforming original_op only if they are in the
          subgraph, otherwise they are ignored.
        """

        # handlers
        self.transform_op_handler = copy_op_handler
        self.transform_control_input_handler = transform_op_if_inside_handler
        self.assign_collections_handler = assign_renamed_collections_handler
        self.transform_external_input_handler = replace_t_with_placeholder_handler
        self.transform_external_hidden_input_handler = keep_t_if_possible_handler
        self.transform_original_op_handler = transform_op_if_inside_handler

    def __call__(self, sgv, dst_graph, dst_scope, src_scope="", reuse_dst_scope=False):
        """Execute the transformation.

        Args:
          sgv: the source subgraph-view.
          dst_graph: the destination graph.
          dst_scope: the destination scope.
          src_scope: the source scope, which specify the path from which the
            relative path of the transformed nodes are computed. For instance, if
            src_scope is a/ and dst_scoped is b/, then the node a/x/y will have a
            relative path of x/y and will be transformed into b/x/y.
          reuse_dst_scope: if True the dst_scope is re-used if it already exists.
            Otherwise, the scope is given a unique name based on the one given
            by appending an underscore followed by a digit (default).
        Returns:
          A tuple `(sgv, info)` where:
            `sgv` is the transformed subgraph view;
            `info` is an instance of TransformerInfo containing
            information about the transform, including mapping between
            original and transformed tensors and operations.
        Raises:
          ValueError: if the arguments are invalid.
        """
        sgv = subgraph.make_view(sgv)
        if not isinstance(dst_graph, tf_ops.Graph):
            raise TypeError("Expected a tf.Graph, got: {}".format(type(dst_graph)))

        src_scope = util.scope_finalize(src_scope)
        dst_scope = util.scope_finalize(dst_scope)

        # Potentially create new scope if reuse_dst_scope is False
        if dst_scope and not reuse_dst_scope:
            dst_scope = util.scope_finalize(dst_graph.unique_name(dst_scope[:-1]))

        # Create temporary info used during this transform call
        info = _TmpInfo(sgv, dst_graph, dst_scope, src_scope)

        self._copy_ops(info)
        self._finalize_cycles(info)
        self._connect_control_inputs(info)

        # Compute information about the transformation
        res_info = TransformerInfo(info)
        sgv_ = self._transform_sgv(info, sgv)
        return sgv_, res_info

    def _copy_ops(self, info):
        """Copy ops without connecting them."""
        sorted_ops = sorted(info.sgv.ops, key=lambda op: op._id)  # pylint: disable=protected-access
        for op in sorted_ops:
            new_inputs = [self._transformed_t(info, t, op) for t in op.inputs]
            op_, op_outputs_ = self.transform_op_handler(info, op, new_inputs)
            if op is op_:
                raise ValueError("In-place transformation not allowed.")

            # Process op.
            info.transformed_ops[op] = op_
            self.assign_collections_handler(info, op, op_)

            # Process output tensors.
            for op_output, op_output_ in zip(op.outputs, op_outputs_):
                info.transformed_ts[op_output] = op_output_
                self.assign_collections_handler(info, op_output, op_output_)

    def _finalize_cycles(self, info):
        """Reconnects the cyclic tensors."""
        for t, tmp_t_, consumer_op in info.tmp_cyclic_ts:
            if t not in info.transformed_ts:
                raise ValueError("The tensor {} should be transformed by now.".format(t.name))
            if consumer_op not in info.transformed_ops:
                raise ValueError("The op {} should be transformed by now.".format(consumer_op.name))
            t_ = info.transformed_ts[t]
            consumer_op_ = info.transformed_ops[consumer_op]
            t_index_ = list(consumer_op_.inputs).index(tmp_t_)
            consumer_op_._update_input(t_index_, t_)  # pylint: disable=protected-access

    def _connect_control_inputs(self, info):
        """Connect the previously copied ops."""
        for op in info.sgv.ops:
            logging.debug("Connecting control inputs of op: %s", op.name)
            op_ = info.transformed_ops[op]

            # Finalize original op.
            # TODO(fkp): Stop worrying about _original_op and remove this code?
            # pylint: disable=protected-access
            if op._original_op:
                original_op = self.transform_original_op_handler(info, op._original_op)
                if original_op is None:
                    logging.debug("Could not find original op for: %s", op_.name)
                else:
                    op_._original_op = original_op
            # pylint: enable=protected-access

            # Finalize control inputs:
            control_inputs_ = [self.transform_control_input_handler(info, ci) for ci in op.control_inputs]
            control_inputs_ = [ci for ci in control_inputs_ if ci is not None]
            reroute.add_control_inputs(op_, control_inputs_)

    def _transform_sgv(self, info, sgv):
        """Transform a subgraph view.

        For convenience, a transform operation returns a subgraph view of the
        transformed graph.

        Args:
          info: Temporary information for this transorfm call.
          sgv: the subgraph to be transformed.
        Returns:
          The transformed subgraph.
        """
        ops_ = [op_ for _, op_ in iteritems(info.transformed_ops)]
        sgv_ = subgraph.SubGraphView(ops_)
        sgv_inputs_ = sgv_.inputs
        sgv_outputs_ = sgv_.outputs

        # re-order inputs
        input_map_ = []
        for input_t in sgv.inputs:
            if input_t not in info.transformed_ts:
                continue
            input_t_ = info.transformed_ts[input_t]
            if input_t_ not in sgv_inputs_:
                continue
            input_t_index_ = sgv_.input_index(input_t_)
            input_map_.append(input_t_index_)

        # re-order outputs
        output_map_ = []
        for output_t in sgv.outputs:
            if output_t not in info.transformed_ts:
                continue
            output_t_ = info.transformed_ts[output_t]
            if output_t_ not in sgv_outputs_:
                continue
            output_t_index_ = sgv_.output_index(output_t_)
            output_map_.append(output_t_index_)

        return sgv_.remap(input_map_, output_map_)

    def _transformed_t(self, info, t, consumer_op):
        """Return tre transformed tensor of `t`."""
        if t in info.transformed_ts:
            # If op is in the subgraph, just return its transformed counterpart.
            return info.transformed_ts[t]

        if t in info.sgv_inputs_set:
            # `t` is an input of the subgraph.
            return self.transform_external_input_handler(info, t)
        elif t.op in info.ops:
            # `t` is an internal tensor but is not transformed yet because it
            # belongs to a graph cycle.
            logging.debug("Cyclic tensor: t.name = %s", t.name)
            # Try to find an existing tensor we can use for now,
            # otherwise create one. We'll rewire this later.
            if consumer_op.type == "Merge":
                first_input = consumer_op.inputs[0]
                tmp_t_ = self._transformed_t(info, first_input, consumer_op)
            elif t.op.type == "Enter":
                enter_input = t.op.inputs[0]
                tmp_t_ = self._transformed_t(info, enter_input, consumer_op)
            else:
                with info.graph_.as_default():
                    tmp_t_ = util.make_placeholder_from_tensor(t, scope=info.scope_, prefix="geph_tmp")
                logging.debug("Created temporary placeholder: %s.", tmp_t_.name)
            # Register as temporary and return.
            info.tmp_cyclic_ts.append((t, tmp_t_, consumer_op))
            return tmp_t_
        else:
            # `t` is a hidden input of the subgraph.
            return self.transform_external_hidden_input_handler(info, t)


def copy(sgv, dst_graph=None, dst_scope="", src_scope="", reuse_dst_scope=False):
    """Copy a subgraph.

    Args:
      sgv: the source subgraph-view. This argument is converted to a subgraph
        using the same rules than the function subgraph.make_view.
      dst_graph: the destination graph.
      dst_scope: the destination scope.
      src_scope: the source scope.
      reuse_dst_scope: if True the dst_scope is re-used if it already exists.
        Otherwise, the scope is given a unique name based on the one given
        by appending an underscore followed by a digit (default).
    Returns:
      A tuple `(sgv, info)` where:
        `sgv` is the transformed subgraph view;
        `info` is an instance of TransformerInfo containing
        information about the transform, including mapping between
        original and transformed tensors and operations.
    Raises:
      TypeError: if `dst_graph` is not a `tf.Graph`.
      StandardError: if sgv cannot be converted to a SubGraphView using
        the same rules than the function subgraph.make_view.
    """
    sgv = subgraph.make_view(sgv)
    if dst_graph is None:
        dst_graph = sgv.graph
    if not isinstance(dst_graph, tf_ops.Graph):
        raise TypeError("Expected a tf.Graph, got: {}".format(type(dst_graph)))

    copier = Transformer()
    return copier(sgv, dst_graph, dst_scope, src_scope, reuse_dst_scope=reuse_dst_scope)


def copy_with_input_replacements(
    sgv, replacement_ts, dst_graph=None, dst_scope="", src_scope="", reuse_dst_scope=False
):
    """Copy a subgraph, replacing some of its inputs.

    Note a replacement only happens if the tensor to be replaced
    is an input of the given subgraph. The inputs of a subgraph can
    be queried using sgv.inputs.

    Args:
      sgv: the source subgraph-view. This argument is converted to a subgraph
        using the same rules as the function subgraph.make_view.
      replacement_ts: dictionary mapping from original tensors to the
        replaced one.
      dst_graph: the destination graph.
      dst_scope: the destination scope.
      src_scope: the source scope.
      reuse_dst_scope: if True the dst_scope is re-used if it already exists.
        Otherwise, the scope is given a unique name based on the one given
        by appending an underscore followed by a digit (default).
    Returns:
      A tuple `(sgv, info)` where:
        `sgv` is the transformed subgraph view;
        `info` is an instance of TransformerInfo containing
        information about the transform, including mapping between
        original and transformed tensors and operations.
    Raises:
      TypeError: if dst_graph is not a tf.Graph.
      StandardError: if sgv cannot be converted to a SubGraphView using
        the same rules as the function subgraph.make_view.
    """
    sgv = subgraph.make_view(sgv)
    if dst_graph is None:
        dst_graph = sgv.graph
    if not isinstance(dst_graph, tf_ops.Graph):
        raise TypeError("Expected a tf.Graph, got: {}".format(type(dst_graph)))

    copier = Transformer()

    # Replace tensor if possible.
    def replace_t_with_replacement_handler(info, t):
        if t in replacement_ts:
            return replacement_ts[t]
        else:
            return keep_t_if_possible_handler(info, t)

    copier.transform_external_input_handler = replace_t_with_replacement_handler
    return copier(sgv, dst_graph, dst_scope, src_scope, reuse_dst_scope=reuse_dst_scope)


def _add_control_flow_ops(ops, control_ios):
    """Complete `ops` so that the transformed graph is valid.

    Partially copying a graph can lead to a malformed graph. For instance,
    copying half of a while construct is likely to result in an invalid graph.
    This function attempts to add missing ops so that the transformation result
    in a valid graph.

    Args:
      ops: list of ops (modifed in-place).
      control_ios: object created by a call to `util.ControlOutputs`.
    """
    # Find while contexts.
    control_flow_contexts = set()
    for op in ops:
        cfc = op._control_flow_context  # pylint: disable=protected-access
        if cfc:
            control_flow_contexts.add(cfc)
    # Find new ops.
    new_ops = []
    for cfc in control_flow_contexts:
        if cfc.IsWhileContext():
            new_ops += select.get_walks_intersection_ops(
                [enter_t.op for enter_t in cfc.loop_enters],
                [exit_t.op for exit_t in cfc.loop_exits],
                control_ios=control_ios,
            )
    # Add new ops.
    new_ops_set = set(new_ops)
    ops_set = frozenset(ops)
    for op in new_ops_set:
        if op not in ops_set:
            ops.append(op)


def graph_replace(target_ts, replacement_ts, dst_scope="", src_scope="", reuse_dst_scope=False):
    """Create a new graph which compute the targets from the replaced Tensors.

    Args:
      target_ts: a single tf.Tensor or an iterable of tf.Tensor.
      replacement_ts: dictionary mapping from original tensors to replaced tensors
      dst_scope: the destination scope.
      src_scope: the source scope.
      reuse_dst_scope: if True the dst_scope is re-used if it already exists.
        Otherwise, the scope is given a unique name based on the one given
        by appending an underscore followed by a digit (default).
    Returns:
      A single tf.Tensor or a list of target tf.Tensor, depending on
      the type of the input argument `target_ts`.
      The returned tensors are recomputed using the tensors from replacement_ts.
    Raises:
      ValueError: if the targets are not connected to replacement_ts.
    """
    # Identify operations in the graph that will change.
    # Start forward walk at Tensors that will be replaced, and
    # backward walk at the target output Tensors.
    flatten_target_ts = util.flatten_tree(target_ts)
    # Construct the forward control dependencies edges so that
    # the get_walks_intersection_ops can also traverse the
    # control dependencies.
    graph = util.get_unique_graph(flatten_target_ts, check_types=(tf.Tensor))
    control_ios = util.ControlOutputs(graph)
    ops = select.get_walks_intersection_ops(list(replacement_ts), flatten_target_ts, control_ios=control_ios)
    if not ops:
        raise ValueError("Targets and replacements are not connected!")

    # Complete ops to avoid malformed control flow.
    # TODO(fkp): Consider moving this function deeper (in the transformer?).
    _add_control_flow_ops(ops, control_ios)

    # Create a copy of the relevant subgraph
    unused_sgv_, info = copy_with_input_replacements(ops, replacement_ts, None, dst_scope, src_scope, reuse_dst_scope)
    # Return the transformed targets but keep the original if the transformed
    # counterpart cannot be found
    missing_fn = lambda original_t: original_t
    return info.transformed(target_ts, missing_fn)



"""
Graph utils.

They might be just available for graph-based frameworks,
but not necessarily.
"""

from __future__ import annotations
from typing import Union, Sequence, Callable, List
from returnn.tensor import Tensor


__all__ = [
    "get_tensor_dependencies",
    "get_tensor_consumers",
    "walk_tensor_consumers",
]


def get_tensor_dependencies(x: Tensor) -> Sequence[Tensor]:
    """
    :param x: tensor
    :return: list of tensors which are inputs to x
    """
    # noinspection PyProtectedMember
    return x._raw_backend.get_tensor_dependencies(x)


def get_tensor_consumers(x: Tensor) -> Sequence[Tensor]:
    """
    :param x: tensor
    :return: list of tensors which consume x
    """
    # noinspection PyProtectedMember
    return x._raw_backend.get_tensor_consumers(x)


def walk_tensor_consumers(
    seed: Union[Tensor, Sequence[Tensor]],
    *,
    filter_outputs: Callable[[Tensor], bool] = None,
    ending_condition: Callable[[Tensor], bool] = None,
) -> List[Tensor]:
    """
    :param seed: tensor
    :param filter_outputs: if given, this function will be called with each tensor,
        and if it returns False, the tensor will be skipped.
    :param ending_condition: if given, this function will be called with each tensor,
        and if it returns True, we return.
    :return: yields all tensors which consume seed, and their consumers, recursively
    """
    seen = set()
    queue = [seed] if isinstance(seed, Tensor) else list(seed)
    outputs = []
    while queue:
        x = queue.pop(0)
        if x in seen:
            continue
        seen.add(x)
        if not filter_outputs or filter_outputs(x):
            outputs.append(x)
        if ending_condition and ending_condition(x):
            break
        queue.extend(get_tensor_consumers(x))
    return outputs



"""
Here we provide some logic to perform hyper-parameter search.
See ``demos/demo-hyper-param-tuning.config`` for an example config.
For each entry in the config where search should be performed on,
you declare it as an instance of :class:`HyperParam`.
Then, this module will find all such instances in the config and replace it with values during search.

The search itself is some evolutionary genetic search.
There are many variants to it, e.g. such as what kind of manipulations you,
e.g. cross-over and mutation, and also, how you sample new random values.
The current logic probably can be improved.

Currently, each search is a training started from the scratch, and the accumulated train score
is used as an evaluation measure.
This probably also can be improved.
Also, instead of always starting from scratch, we could keep intermediate results and resume from them,
or use real training intermediate results and resume from them.
We could even do some simple search in the beginning of each epoch when we keep it cheap enough.

Also, we could store the population of hyper params on disk to allow resuming of a search.
"""

from __future__ import annotations

import sys
import time
import typing
import numpy
import returnn.tf.compat as tf_compat
from returnn.config import Config
from returnn.log import log
from returnn.datasets import Dataset
from returnn.datasets.generating import StaticDataset
from returnn.tf.engine import Engine, Runner, CancelTrainingException
from returnn.util.basic import CollectionReadCheckCovered, hms_fraction, guess_requested_max_num_threads


Eps = 1e-16


class HyperParam:
    """
    Represents one hyper parameter.
    """

    # noinspection PyShadowingNames
    def __init__(self, dtype=None, bounds=None, classes=None, log=False, default=None):
        """
        :param str|type|None|list dtype: e.g. "float", "int" or "bool", or if Collection, will be classes
        :param None|list[int|float] bounds: inclusive
        :param list|None classes:
        :param bool log: if in log-scale
        :param float|int|object|None default:
        """
        if isinstance(dtype, (list, tuple)):
            assert classes is None
            assert bounds is None
            classes = dtype
            dtype = None
        if dtype is None:
            assert classes is not None
        elif dtype == "float":
            dtype = float
        elif dtype == "int":
            dtype = int
        elif dtype == "bool":
            dtype = bool
        assert dtype in (float, int, bool, None)
        if bounds is not None:
            assert dtype in (int, float)
            assert isinstance(bounds, (list, tuple))
            assert len(bounds) == 2
            assert dtype(bounds[0]) < dtype(bounds[1])
        if classes is not None:
            assert isinstance(classes, (list, tuple)), "should be with a defined order"
            assert len(classes) > 0
        self.dtype = dtype
        self.bounds = bounds
        self.classes = classes
        self.log_space = log
        self.default = default
        self.unique_idx = HyperParam._get_next_unique_idx()
        self.usages = []  # type: typing.List[_AttrChain]

    _unique_idx = 0

    @classmethod
    def _get_next_unique_idx(cls):
        cls._unique_idx += 1
        return cls._unique_idx

    def __repr__(self):
        if self.classes is not None:
            return "HyperParam(%r)" % self.classes
        dtype_name = self.dtype.__name__
        ext = ""
        if self.log_space:
            ext += ", log=True"
        if self.default is not None:
            ext += ", default=%r" % self.default
        if self.bounds is not None:
            return "HyperParam(%s, %s%s)" % (dtype_name, self.bounds, ext)
        assert self.bounds is None
        return "HyperParam(%s%s)" % (dtype_name, ext)

    def get_canonical_usage(self):
        """
        :rtype: _AttrChain
        """
        return self.get_sorted_usages()[0]

    def get_sorted_usages(self):
        """
        :rtype: list[_AttrChain]
        """
        return sorted(self.usages, key=lambda chain: min(2, len(chain.chain)))

    def description(self):
        """
        :rtype: str
        """
        if len(self.usages) == 0:
            usage_str = "<no usage>"
        elif len(self.usages) == 1:
            usage_str = str(self.usages[0])
        else:
            usage_str = str(self.get_canonical_usage()) + "|..."
        return usage_str + ": %s" % self

    def get_num_instances(self, upper_limit=100):
        """
        :param int upper_limit:
        :rtype: int
        """
        assert upper_limit >= 2
        if self.classes is not None:
            return min(len(self.classes), upper_limit)
        if self.dtype is bool:
            return 2
        if self.dtype is float:
            return upper_limit
        if self.dtype is int:
            x1, x2 = self.bounds
            x1 = numpy.ceil(x1)
            x2 = numpy.floor(x2)
            assert x1 < x2
            return min(x2 - x1 + 1, upper_limit)
        raise Exception("invalid dtype %r" % self.dtype)

    def merge_values(self, value1, value2):
        """
        Merge two values, which are valid values for this `HyperParam`.

        :param T value1:
        :param T value2:
        :rtype: T
        """
        if self.dtype is bool:
            return value1
        if self.log_space:
            x0, x1 = value1, value2
            if x0 > x1:
                x0, x1 = x1, x0
            if x0 < 0 or x1 < 0:
                assert x0 <= x1 <= 0
                sign = -1
                x0, x1 = -x1, -x0
            else:
                sign = 1
            assert x1 >= x0 >= 0
            x0o = x0
            if x0 < Eps * 0.5:
                x0 = Eps * 0.5
            if x1 < Eps:
                x1 = Eps
            x0 = numpy.log(float(x0))
            x1 = numpy.log(float(x1))
            y = numpy.exp(x0 + (x1 - x0) * 0.5)
            if y <= Eps:
                y = x0o
            return self.dtype(y) * sign
        if self.dtype is int:
            return (value1 + value2) // 2
        return self.dtype((value1 + value2) * 0.5)

    def get_value(self, selected, eps=Eps):
        """
        :param float selected: must be between 0 and 1
        :param float eps: if in log-space and you have e.g. bounds=[0,1], will be the lowest value, before 0. see code.
        :rtype: float|int|bool|object
        """
        assert 0 < eps
        assert 0 <= selected <= 1
        if self.classes:
            return self.classes[int(len(self.classes) * selected)]
        if self.dtype is bool:
            return selected > 0.5
        if self.bounds:
            if self.dtype is int and not self.log_space:
                return self.bounds[0] + int((self.bounds[1] - self.bounds[0]) * selected)
            if self.log_space:
                x0, x1 = self.bounds
                if x0 < 0 or x1 < 0:
                    assert x0 < x1 <= 0
                    sign = -1
                    x0, x1 = -x1, -x0
                else:
                    sign = 1
                assert x1 > x0 >= 0
                x0b, x1b = x0, x1
                if x0b < eps * 0.5:
                    x0b = eps * 0.5
                if x1b < eps:
                    x1b = eps
                x0l = numpy.log(float(x0b))
                x1l = numpy.log(float(x1b))
                y = numpy.exp(x0l + (x1l - x0l) * selected)
                if y <= eps:
                    y = x0
                return self.dtype(y) * sign
            return self.dtype(self.bounds[0] + (self.bounds[1] - self.bounds[0]) * selected)
        # No bounds. So anything -inf to inf.
        # But exclude -inf/inf.
        # Assume selected is uniform in [0,1], so use the inverse accumulated Gauss density function
        # to get normal distributed in [-inf,inf].
        x = selected
        if x < eps:
            x = eps
        if x > 1.0 - eps:
            x = 1.0 - eps
        import scipy.special

        return self.dtype(scipy.special.ndtri(x))  # noqa

    def get_initial_value(self):
        """
        :rtype: float|int|bool|object
        """
        return self.get_value(selected=0.5)

    def get_default_value(self):
        """
        :rtype: float|int|bool|object
        """
        if self.default is not None:
            return self.dtype(self.default)
        return self.get_initial_value()

    def get_random_value(self, seed, eps=Eps):
        """
        :param int seed:
        :param float eps: see get_value()
        :rtype: float|int|bool|object
        """
        rnd = numpy.random.RandomState(seed=seed)
        x = rnd.uniform(0.0, 1.0)
        if x < eps:
            x = 0.0
        if x > 1.0 - eps:
            x = 1.0
        return self.get_value(x, eps=eps)

    def get_random_value_by_idx(self, iteration_idx, individual_idx):
        """
        :param int iteration_idx:
        :param int individual_idx:
        :rtype: float|int|bool|object
        """
        # Use a deterministic seed for the random number generator
        # which will not change on unrelated changes in the config file,
        # so that runs will stay deterministic in this sense.
        seed = hash_obj((self.get_canonical_usage(), iteration_idx, individual_idx))
        return self.get_random_value(seed=seed)


class TrainException(Exception):
    """
    Exception from training.
    """


class Individual:
    """
    One instance of hyper params.
    """

    def __init__(self, hyper_param_mapping, name):
        """
        :param dict[HyperParam] hyper_param_mapping:
        :param str name:
        """
        self.hyper_param_mapping = hyper_param_mapping
        self.cost = None
        self.name = name

    def cross_over(self, hyper_params, population, random_seed):
        """
        :param list[HyperParam] hyper_params:
        :param list[Individual] population:
        :param int random_seed:
        :return: copy of self, cross-overd with others
        :rtype: Individual
        """
        name = self.name
        if len(name) > 10:
            name = name[:8] + ".."
        name += "x%x" % random_seed
        res = Individual(hyper_param_mapping=self.hyper_param_mapping.copy(), name=name)
        rnd = numpy.random.RandomState(random_seed)
        while True:
            other = population[rnd.random_integers(0, len(population) - 1)]
            for p in hyper_params:
                x = rnd.uniform(0.0, 1.0)
                if x > 0.75:
                    res.hyper_param_mapping[p] = other.hyper_param_mapping[p]
                elif x > 0.5:
                    res.hyper_param_mapping[p] = p.merge_values(
                        res.hyper_param_mapping[p], other.hyper_param_mapping[p]
                    )
            if rnd.uniform(0.0, 1.0) > 0.5:
                break
        return res


class Optimization:
    """
    Hyper parameter optimization handler class.
    """

    def __init__(self, config, train_data):
        """
        :param returnn.config.Config config:
        :param Dataset train_data:
        """
        self.config = config
        self.opts = CollectionReadCheckCovered(config.get_of_type("hyper_param_tuning", dict, {}))
        self.log = log.v1
        train_data.init_seq_order(epoch=1)
        self.train_data = StaticDataset.copy_from_dataset(train_data, max_seqs=self.opts.get("num_train_steps", 100))
        self.hyper_params = []  # type: typing.List[HyperParam]
        self._find_hyper_params()
        if not self.hyper_params:
            raise Exception("No hyper params found.")
        self.hyper_params.sort(key=lambda p_: p_.unique_idx)
        print("We have found these hyper params:")
        for p in self.hyper_params:
            print(" %s" % p.description())
        self.dry_run_first_individual = self.opts.get("dry_run_first_individual", True)
        self.num_iterations = self.opts["num_tune_iterations"]
        self.num_individuals = self.opts["num_individuals"]
        self.num_kill_individuals = self.opts.get("num_kill_individuals", self.num_individuals // 2)
        self.num_best = self.opts.get("num_best", 10)
        self.num_threads = self.opts.get("num_threads", guess_requested_max_num_threads())
        self.opts.assert_all_read()

    def _find_hyper_params(self, base=None, visited=None):
        """
        :param _AttrChain base:
        :param set[int] visited: set of ids
        """
        from inspect import ismodule

        if base is None:
            base = _AttrChain(base=self.config)
        if isinstance(base.value, HyperParam):
            base.value.usages.append(base)
            if base.value not in self.hyper_params:
                self.hyper_params.append(base.value)
            return
        if visited is None:
            visited = set()
        if id(base.value) in visited:
            return
        visited.add(id(base.value))
        if ismodule(base.value):
            return
        if isinstance(base.value, dict):
            col_type = _AttribOrKey.ColTypeDict
            keys = base.value.keys()
        elif isinstance(base.value, Config):
            col_type = _AttribOrKey.ColTypeConfig
            keys = base.value.typed_dict.keys()
        else:
            # Add other specific object types, but not in generic all.
            return
        for key in sorted(keys):
            child = base.get_extended_chain(_AttribOrKey(key=key, col_type=col_type))
            self._find_hyper_params(base=child, visited=visited)

    def get_population(self, iteration_idx, num_individuals):
        """
        :param int iteration_idx:
        :param int num_individuals:
        :rtype: list[Individual]
        """
        assert num_individuals > 0
        return [self.get_individual(iteration_idx=iteration_idx, individual_idx=i) for i in range(num_individuals)]

    def get_individual(self, iteration_idx, individual_idx):
        """
        :param int iteration_idx:
        :param int individual_idx:
        :rtype: Individual
        """
        return Individual(
            {
                p: p.get_random_value_by_idx(iteration_idx=iteration_idx, individual_idx=individual_idx)
                for p in self.hyper_params
            },
            name="%i-%i" % (iteration_idx, individual_idx),
        )

    def cross_over(self, population, iteration_idx):
        """
        :param list[Individual] population: modified in-place
        :param int iteration_idx:
        """
        for i in range(len(population) - 1):
            population[i] = population[i].cross_over(
                hyper_params=self.hyper_params,
                population=population[:i] + population[i + 1 :],
                random_seed=iteration_idx * 1013 + i * 17,
            )

    def create_config_instance(self, hyper_param_mapping, gpu_ids):
        """
        :param dict[HyperParam] hyper_param_mapping: maps each hyper param to some value
        :param set[int] gpu_ids:
        :rtype: Config
        """
        assert set(self.hyper_params) == set(hyper_param_mapping.keys())
        from returnn.util.basic import deepcopy

        config = deepcopy(self.config)
        assert isinstance(config, Config)
        for p, value in hyper_param_mapping.items():
            assert isinstance(p, HyperParam)
            for attr_chain in p.usages:
                attr_chain.write_attrib(base=config, new_value=value)
        tf_session_opts = config.typed_dict.setdefault("tf_session_opts", {})
        # https://github.com/tensorflow/tensorflow/blob/master/tensorflow/core/protobuf/config.proto
        gpu_opts = tf_session_opts.setdefault("gpu_options", tf_compat.v1.GPUOptions())
        if isinstance(gpu_opts, dict):
            gpu_opts = tf_compat.v1.GPUOptions(**gpu_opts)
        gpu_opts.visible_device_list = ",".join(map(str, sorted(gpu_ids)))
        return config

    def work(self):
        """
        Start the optimization.
        """
        print("Starting hyper param search. Using %i threads." % self.num_threads, file=log.v1)
        from returnn.tf.util.basic import get_available_gpu_devices
        from returnn.log import wrap_log_streams, StreamDummy
        from threading import Thread, Condition
        from returnn.util.basic import progress_bar, hms, is_tty

        class Outstanding:
            """
            Queue of outstanding work.
            """

            cond = Condition()
            threads = []  # type: typing.List[WorkerThread]
            population = []
            exit = False
            exception = None

        class WorkerThread(Thread):
            """
            Worker threader.
            """

            def __init__(self, gpu_ids):
                """
                :param set[int] gpu_ids:
                """
                super(WorkerThread, self).__init__(name="Hyper param tune train thread")
                self.gpu_ids = gpu_ids
                self.trainer = None  # type: typing.Optional[_IndividualTrainer]
                self.finished = False
                self.start()

            def cancel(self, join=False):
                """
                :param bool join:
                """
                with Outstanding.cond:
                    if self.trainer:
                        self.trainer.cancel_flag = True
                        if self.trainer.runner:
                            self.trainer.runner.cancel_flag = True
                if join:
                    self.join()

            def get_complete_frac(self):
                """
                :rtype: float
                """
                with Outstanding.cond:
                    if self.trainer and self.trainer.runner:
                        return self.trainer.runner.data_provider.get_complete_frac()
                return 0.0

            # noinspection PyMethodParameters
            def run(self_thread):
                """
                Run thread.
                """
                try:
                    while True:
                        with Outstanding.cond:
                            if Outstanding.exit or Outstanding.exception:
                                return
                            if not Outstanding.population:
                                self_thread.finished = True
                                Outstanding.cond.notify_all()
                                return
                            # noinspection PyShadowingNames
                            individual = Outstanding.population.pop(0)
                            self_thread.trainer = _IndividualTrainer(
                                optim=self, individual=individual, gpu_ids=self_thread.gpu_ids
                            )
                        self_thread.name = "Hyper param tune train thread on %r" % individual.name
                        self_thread.trainer.run()
                except Exception as exc:
                    with Outstanding.cond:
                        if not Outstanding.exception:
                            Outstanding.exception = exc or True
                        Outstanding.cond.notify_all()
                    for thread in Outstanding.threads:
                        if thread is not self_thread:
                            thread.cancel()
                    if not isinstance(exc, CancelTrainingException):
                        with Outstanding.cond:  # So that we don't mix up multiple on sys.stderr.
                            # This would normally dump it on sys.stderr so it's fine.
                            sys.excepthook(*sys.exc_info())

        best_individuals = []
        population = []
        num_gpus = len(get_available_gpu_devices())
        print("Num available GPUs:", num_gpus)
        num_gpus = num_gpus or 1  # Would be ignored anyway.
        interactive = is_tty()
        try:
            print(
                "Population of %i individuals (hyper param setting instances), running for %i evaluation iterations."
                % (self.num_individuals, self.num_iterations),
                file=log.v2,
            )
            for cur_iteration_idx in range(1, self.num_iterations + 1):
                print("Starting iteration %i." % cur_iteration_idx, file=log.v2)
                if cur_iteration_idx == 1:
                    population.append(Individual({p: p.get_default_value() for p in self.hyper_params}, name="default"))
                    population.append(
                        Individual({p: p.get_initial_value() for p in self.hyper_params}, name="canonical")
                    )
                population.extend(
                    self.get_population(
                        iteration_idx=cur_iteration_idx, num_individuals=self.num_individuals - len(population)
                    )
                )
                if cur_iteration_idx > 1:
                    self.cross_over(population=population, iteration_idx=cur_iteration_idx)
                if cur_iteration_idx == 1 and self.dry_run_first_individual:
                    # Train first directly for testing and to see log output.
                    # Later we will strip away all log output.
                    print("Very first try with log output:", file=log.v2)
                    _IndividualTrainer(optim=self, individual=population[0], gpu_ids={0}).run()
                print("Starting training with thread pool of %i threads." % self.num_threads)
                iteration_start_time = time.time()
                with wrap_log_streams(StreamDummy(), also_sys_stdout=True, tf_log_verbosity="WARN"):
                    Outstanding.exit = False
                    Outstanding.population = list(population)
                    Outstanding.threads = [WorkerThread(gpu_ids={i % num_gpus}) for i in range(self.num_threads)]
                    try:
                        while True:
                            with Outstanding.cond:
                                if all([thread.finished for thread in Outstanding.threads]) or Outstanding.exception:
                                    break
                                complete_frac = max(
                                    len(population) - len(Outstanding.population) - len(Outstanding.threads), 0
                                )
                                complete_frac += sum([thread.get_complete_frac() for thread in Outstanding.threads])
                                complete_frac /= float(len(population))
                                remaining_str = ""
                                if complete_frac > 0:
                                    start_elapsed = time.time() - iteration_start_time
                                    total_time_estimated = start_elapsed / complete_frac
                                    remaining_estimated = total_time_estimated - start_elapsed
                                    remaining_str = hms(remaining_estimated)
                                if interactive:
                                    progress_bar(complete_frac, prefix=remaining_str, file=sys.__stdout__)
                                else:
                                    print(
                                        "Progress: %.02f%%" % (complete_frac * 100),
                                        "remaining:",
                                        remaining_str or "unknown",
                                        file=sys.__stdout__,
                                    )
                                    sys.__stdout__.flush()
                                Outstanding.cond.wait(1 if interactive else 10)
                        for thread in Outstanding.threads:
                            thread.join()
                    finally:
                        Outstanding.exit = True
                        for thread in Outstanding.threads:
                            thread.cancel(join=True)
                Outstanding.threads = []
                print("Training iteration elapsed time:", hms(time.time() - iteration_start_time))
                if Outstanding.exception:
                    raise Outstanding.exception
                assert not Outstanding.population
                print("Training iteration finished.")
                population.sort(key=lambda p: p.cost)
                del population[-self.num_kill_individuals :]
                best_individuals.extend(population)
                best_individuals.sort(key=lambda p: p.cost)
                del best_individuals[self.num_best :]
                population = best_individuals[: self.num_kill_individuals // 4] + population
                print(
                    "Current best setting, individual %s" % best_individuals[0].name, "cost:", best_individuals[0].cost
                )
                for p in self.hyper_params:
                    print(" %s -> %s" % (p.description(), best_individuals[0].hyper_param_mapping[p]))
        except KeyboardInterrupt:
            print("KeyboardInterrupt, canceled search.")

        print("Best %i settings:" % len(best_individuals))
        for individual in best_individuals:
            print("Individual %s" % individual.name, "cost:", individual.cost)
            for p in self.hyper_params:
                print(" %s -> %s" % (p.description(), individual.hyper_param_mapping[p]))


class _IndividualTrainer:
    def __init__(self, optim, individual, gpu_ids):
        """
        :param Optimization optim:
        :param Individual individual:
        :param set[int] gpu_ids:
        """
        self.optim = optim
        self.individual = individual
        self.runner = None  # type: typing.Optional[Runner]
        self.gpu_ids = gpu_ids
        self.cancel_flag = False

    def run(self):
        """
        Run the trainer.
        """
        if self.individual.cost is not None:
            return self.individual.cost
        start_time = time.time()
        hyper_param_mapping = self.individual.hyper_param_mapping
        print("Training %r using hyper params:" % self.individual.name, file=log.v2)
        for p in self.optim.hyper_params:
            print(" %s -> %s" % (p.description(), hyper_param_mapping[p]), file=log.v2)
        config = self.optim.create_config_instance(hyper_param_mapping, gpu_ids=self.gpu_ids)
        # init_train_from_config expects to have the train task
        config.set("task", "train")
        engine = Engine(config=config)
        train_data = StaticDataset.copy_from_dataset(self.optim.train_data)
        engine.init_train_from_config(config=config, train_data=train_data)
        # Not directly calling train() as we want to have full control.
        engine.epoch = 1
        train_data.init_seq_order(epoch=engine.epoch)
        batches = train_data.generate_batches(
            recurrent_net=engine.network.recurrent,
            batch_size=engine.batch_size,
            max_seqs=engine.max_seqs,
            max_seq_length=int(engine.max_seq_length),
            seq_drop=engine.seq_drop,
            shuffle_batches=engine.shuffle_batches,
            used_data_keys=engine.network.used_data_keys,
        )
        engine.updater.set_learning_rate(engine.learning_rate, session=engine.tf_session)
        trainer = Runner(engine=engine, dataset=train_data, batches=batches, train=True)
        self.runner = trainer
        if self.cancel_flag:
            raise CancelTrainingException("Trainer cancel flag is set")
        trainer.run(report_prefix="hyper param tune train %r" % self.individual.name)
        if not trainer.finalized:
            print("Trainer exception:", trainer.run_exception, file=log.v1)
            raise trainer.run_exception
        cost = trainer.score["cost:output"]
        print(
            "Individual %s:" % self.individual.name,
            "Train cost:",
            cost,
            "elapsed time:",
            hms_fraction(time.time() - start_time),
            file=self.optim.log,
        )
        self.individual.cost = cost


class _AttribOrKey:
    ColTypeConfig = Config
    ColTypeDict = dict
    ColTypeObj = object

    def __init__(self, key, col_type):
        """
        :param str|object key:
        :param type[object]|type[dict] col_type:
        """
        self.key = key
        self.col_type = col_type

    def __str__(self):
        if self.col_type is self.ColTypeConfig:
            return "%s" % self.key
        if self.col_type is self.ColTypeDict:
            return "[%r]" % self.key
        if self.col_type is self.ColTypeObj:
            return ".%s" % self.key
        raise Exception("invalid col_type %r" % self.col_type)

    def get(self, parent):
        """
        :param object|dict|Config parent:
        :rtype: dict|object|HyperParam
        """
        if self.col_type is self.ColTypeConfig:
            return parent.typed_dict[self.key]
        if self.col_type is self.ColTypeDict:
            return parent[self.key]
        if self.col_type is self.ColTypeObj:
            return getattr(parent, self.key)
        raise Exception("invalid col_type %r" % self.col_type)

    def set(self, parent, new_value):
        """
        :param object|dict|Config parent:
        :param new_value:
        """
        if self.col_type is self.ColTypeConfig:
            parent.typed_dict[self.key] = new_value
            return
        if self.col_type is self.ColTypeDict:
            parent[self.key] = new_value
            return
        if self.col_type is self.ColTypeObj:
            setattr(parent, self.key, new_value)
            return
        raise Exception("invalid col_type %r" % self.col_type)


class _AttrChain:
    def __init__(self, base):
        """
        :param object|dict base:
        """
        self.base = base
        self.chain = []  # type: typing.List[_AttribOrKey]
        self.value = base  # type: typing.Union[HyperParam,object]

    def __str__(self):
        return "".join(map(str, self.chain))

    def __repr__(self):
        return "<%s %r %r>" % (self.__class__.__name__, self.chain, self.value)

    def get_extended_chain(self, attr):
        """
        :param _AttribOrKey attr:
        :rtype: _AttrChain
        """
        sub_chain = _AttrChain(base=self.base)
        sub_chain.chain = list(self.chain)
        sub_chain.chain.append(attr)
        sub_chain.value = attr.get(self.value)
        return sub_chain

    def write_attrib(self, base, new_value):
        """
        :param object|dict|Config base:
        :param new_value:
        """
        obj = base
        assert len(self.chain) >= 1
        for attr in self.chain[:-1]:
            obj = attr.get(obj)
        self.chain[-1].set(obj, new_value)


def hash_str_djb2(s):
    """
    :param str s:
    :rtype: int
    """
    v = 5381
    for x in s:
        v = ((v << 5) + v) + ord(x)
        v = v & 0xFFFFFFFF
    return v


def hash_seq(ls):
    """
    :param list|tuple ls:
    :rtype: int
    """
    v = 5381
    for x in ls:
        v = 1000003 * v + hash_obj(x)
        v = v & 0xFFFFFFFF
    return v


def hash_int(x):
    """
    :param int x:
    :rtype: int
    """
    return ((x << 11) + x) & 0xFFFFFFFF


def hash_obj(x):
    """
    :param tuple|list|str|_AttribOrKey|_AttrChain x:
    :rtype: int
    """
    if isinstance(x, (list, tuple)):
        return hash_seq(x)
    if isinstance(x, str):
        return hash_str_djb2(x)
    if isinstance(x, _AttribOrKey):
        return hash_str_djb2(x.key)
    if isinstance(x, _AttrChain):
        return hash_seq(x.chain)
    if isinstance(x, int):
        return hash_int(x)
    raise TypeError("invalid type %s" % type(x))



"""setup.py script for warp-rna TensorFlow wrapper
mostly copied from https://github.com/HawkAaron/warp-transducer/blob/master/tensorflow_binding/setup.py
Original License: Apache License 2.0
Changes are basically just adapted to the RNA code (and some formatting changes).
"""

from __future__ import print_function

import os
import platform
import re
import warnings
import setuptools
import sys
import unittest
from setuptools.command.build_ext import build_ext as orig_build_ext
from distutils.version import LooseVersion

# We need to import tensorflow to find where its include directory is.
try:
    import tensorflow as tf
except ImportError:
    raise RuntimeError("Tensorflow must be installed to build the tensorflow wrapper.")

if "CUDA_HOME" not in os.environ:
    print("CUDA_HOME not found in the environment so building "
          "without GPU support. To build with GPU support "
          "please define the CUDA_HOME environment variable. "
          "This should be a path which contains include/cuda.h",
          file=sys.stderr)
    enable_gpu = False
else:
    enable_gpu = True

if platform.system() == 'Darwin':
    lib_ext = ".dylib"
else:
    lib_ext = ".so"

warp_rna_path = "../build"
if "WARP_RNA_PATH" in os.environ:
    warp_rna_path = os.environ["WARP_RNA_PATH"]
if not os.path.exists(os.path.join(warp_rna_path, "libwarprna" + lib_ext)):
    print(("Could not find libwarprna.so in {}.\n"
           "Build warp-rna and set WARP_RNA_PATH to the location of"
           " libwarprna.so (default is '../build')").format(warp_rna_path),
          file=sys.stderr)
    sys.exit(1)

root_path = os.path.realpath(os.path.dirname(__file__))

tf_include = tf.sysconfig.get_include()
tf_src_dir = tf.sysconfig.get_lib()
tf_includes = [tf_include, tf_src_dir]
include_dirs = tf_includes + [os.path.join(root_path, '../include')]

if LooseVersion(tf.__version__) >= LooseVersion('1.4'):
    nsync_dir = '../../external/nsync/public'
    if LooseVersion(tf.__version__) >= LooseVersion('1.10'):
        nsync_dir = 'external/nsync/public'
    include_dirs += [os.path.join(tf_include, nsync_dir)]

if os.getenv("TF_CXX11_ABI") is not None:
    TF_CXX11_ABI = os.getenv("TF_CXX11_ABI")
else:
    warnings.warn("Assuming tensorflow was compiled without C++11 ABI. "
                  "It is generally true if you are using binary pip package. "
                  "If you compiled tensorflow from source with gcc >= 5 and didn't set "
                  "-D_GLIBCXX_USE_CXX11_ABI=0 during compilation, you need to set "
                  "environment variable TF_CXX11_ABI=1 when compiling this bindings. "
                  "Also be sure to touch some files in src to trigger recompilation. "
                  "Also, you need to set (or unsed) this environment variable if getting "
                  "undefined symbol: _ZN10tensorflow... errors")
    TF_CXX11_ABI = "0"

extra_compile_args = ['-std=c++11', '-fPIC', '-D_GLIBCXX_USE_CXX11_ABI=' + TF_CXX11_ABI]
# current tensorflow code triggers return type errors, silence those for now
extra_compile_args += ['-Wno-return-type']
if LooseVersion(tf.__version__) >= LooseVersion('1.4'):
    extra_compile_args += tf.sysconfig.get_compile_flags()

extra_link_args = []
if LooseVersion(tf.__version__) >= LooseVersion('1.4'):
    extra_link_args += tf.sysconfig.get_link_flags()

if (enable_gpu):
    extra_compile_args += ['-DWARPRNA_ENABLE_GPU']
    include_dirs += [os.path.join(os.environ["CUDA_HOME"], 'include')]

    # mimic tensorflow cuda include setup so that their include command work
    if not os.path.exists(os.path.join(root_path, "include")):
        os.mkdir(os.path.join(root_path, "include"))

    cuda_inc_path = os.path.join(root_path, "include/cuda")
    if not os.path.exists(cuda_inc_path) or os.readlink(cuda_inc_path) != os.environ["CUDA_HOME"]:
        if os.path.exists(cuda_inc_path):
            os.remove(cuda_inc_path)
        os.symlink(os.environ["CUDA_HOME"], cuda_inc_path)
    include_dirs += [os.path.join(root_path, 'include')]

# Ensure that all expected files and directories exist.
for loc in include_dirs:
    if not os.path.exists(loc):
        print(("Could not find file or directory {}.\n"
               "Check your environment variables and paths?").format(loc),
              file=sys.stderr)
        sys.exit(1)

lib_srcs = ['src/warp_rna_op.cc']

ext = setuptools.Extension('warprna_tensorflow.kernels',
                           sources = lib_srcs,
                           language = 'c++',
                           include_dirs = include_dirs,
                           library_dirs = [warp_rna_path],
                           runtime_library_dirs = [os.path.realpath(warp_rna_path)],
                           libraries = ['warp_rna'],
                           extra_compile_args = extra_compile_args,
                           extra_link_args = extra_link_args)

class build_tf_ext(orig_build_ext):
    def build_extensions(self):
        if LooseVersion(tf.__version__) < LooseVersion('1.4'):
            self.compiler.compiler_so.remove('-Wstrict-prototypes')
        orig_build_ext.build_extensions(self)

def discover_test_suite():
    test_loader = unittest.TestLoader()
    test_suite = test_loader.discover('tests', pattern='test_*.py')
    return test_suite

# Read the README.md file for the long description. This lets us avoid
# duplicating the package description in multiple places in the source.
README_PATH = os.path.join(os.path.dirname(__file__), "README.md")
with open(README_PATH, "r") as handle:
    # Extract everything between the first set of ## headlines
    LONG_DESCRIPTION = re.search("#.*([^#]*)##", handle.read()).group(1).strip()

setuptools.setup(
    name = "warp_rna",
    version = "0.1",
    description = "TensorFlow wrapper for warp-rna",
    url="https://github.com/1ytic/warp-rna",
    long_description = LONG_DESCRIPTION,
    author = "Mingkun Huang",
    author_email = "mingkunhuang95@gmail.com",
    license = "Apache",
    packages = ["warp_rna"],
    ext_modules = [ext],
    cmdclass = {'build_ext': build_tf_ext},
    test_suite = 'setup.discover_test_suite',
)
